# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Nathan Embery

from typing import List

from skilletlib.snippet.docker import DockerSnippet
from .base import Skillet


class DockerSkillet(Skillet):
    snippet_required_metadata = {'name', 'image', 'cmd'}

    # optional parameters that may be set in the snippet metadata
    snippet_optional_metadata = {
        'volumes': dict(),
        'async': True
    }

    def __init__(self, s: dict):
        # grab the configured 'volumes' from the skillet app_data if present
        # note, this attribute can only be injected by the application and in the skillet definition file
        # also grab the working dir that should be used in the app, this will depend on any volumes that may be
        # passed in as well
        if 'app_data' in s and isinstance(s['app_data'], dict):
            self.volumes = s['app_data'].get('volumes', list())
            self.working_dir = s['app_data'].get('working_dir', '/app')
        else:
            self.volumes = list()
            self.working_dir = None

        super().__init__(s)

    def get_snippets(self) -> List[DockerSnippet]:

        if hasattr(self, 'snippets'):
            return self.snippets

        snippet_list = list()

        for snippet_def in self.snippet_stack:
            # self.path is set automatically in skillet/base.py
            # set skillet_path here for each skillet to have access to current path
            # this is needed for host directory mapping for volume mounts
            snippet_def['skillet_path'] = self.path

            if self.volumes:
                snippet_def['volumes'] = self.volumes
            else:
                snippet_def['volumes'] = list()

            # ensure working dir gets passed in as well...
            # prefer working set on the snippet metadata
            # then prefer
            working_dir = snippet_def.get('working_dir', None)

            if self.working_dir is not None:
                snippet_def['working_dir'] = self.working_dir

            elif working_dir is not None:
                snippet_def['working_dir'] = working_dir

            else:
                snippet_def['working_dir'] = '/app'

            snippet = DockerSnippet(snippet_def)
            snippet_list.append(snippet)

        return snippet_list

    def cleanup(self):
        for snippet in self.snippets:
            snippet.cleanup()



# Copyright (c) 2020, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Adam Baumeister, Nathan Embery

import logging
import os
import shutil

from git import Repo
from git.exc import GitCommandError

from skilletlib.exceptions import SkilletLoaderException

logger = logging.getLogger(__name__)


class Git:
    """
    Git remote
    This class provides an interface to Github repositories containing Skillets or XML snippets.
    """

    def __init__(self, repo_url, store=os.getcwd()):
        """
        Initialize a new Git repo object
        :param repo_url: URL path to repository.
        :param store: Directory to store repository in. Defaults to the current directory.
        """

        if not self.check_git_exists():
            raise SkilletLoaderException('A git client must be installed to use this remote!')

        self.repo_url = repo_url
        self.store = store
        self.Repo = None
        self.name = ''
        self.path = ''
        self.update = ''

    def clone(self, name: str) -> str:
        """
        Clone a remote directory into the store.
        :param name: Name of repository
        :return: (string): Path to cloned repository
        """
        if not name:
            raise ValueError("Missing or bad name passed to Clone command.")

        self.name = name
        path = self.store + os.sep + name
        self.path = path

        if os.path.exists(path):
            self.Repo = Repo(path)

            # FIX for #56
            if self.repo_url not in self.Repo.remotes.origin.urls:
                logger.info('Found new remote URL for this named repo')

                try:
                    # only recourse is to remove the .git directory
                    if os.path.exists(os.path.join(path, '.git')):
                        shutil.rmtree(path)

                    else:
                        raise SkilletLoaderException('Refusing to remove non-git directory')

                except OSError:
                    raise SkilletLoaderException('Repo directory exists!')

                logger.debug("Cloning into {}".format(path))

                try:
                    self.Repo = Repo.clone_from(self.repo_url, path)

                except GitCommandError as gce:
                    raise SkilletLoaderException(f'Could not clone repository {gce}')

            else:
                logger.debug("Updating repository...")

                try:
                    self.Repo.remotes.origin.pull()

                except GitCommandError as gce:
                    logger.error('Could not clone repository!')
                    raise SkilletLoaderException(f'Error Cloning repository {gce}')

            return path

        else:
            logger.debug("Cloning into {}".format(path))
            self.Repo = Repo.clone_from(self.repo_url, path)

        self.path = path
        return path

    def branch(self, branch_name: str) -> None:
        """
        Checkout the specified branch.
        :param branch_name: Branch to checkout.
        :return: None
        """
        logger.debug("Checking out: " + branch_name)

        if self.update:
            logger.debug("Updating branch.")
            self.Repo.remotes.origin.pull()

        self.Repo.git.checkout(branch_name)

    @staticmethod
    def check_git_exists():
        return shutil.which("git")



# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Adam Baumeister, Nathan Embery

import html
import logging
from pathlib import Path
from typing import List
from typing import Optional

import skilletlib
from skilletlib.panoply import Panoply
from skilletlib.snippet.panos import PanosSnippet
from .base import Skillet
from ..exceptions import SkilletLoaderException
from ..exceptions import SkilletValidationException

logger = logging.getLogger(__name__)


class PanosSkillet(Skillet):
    panoply = None

    snippet_required_metadata = {'name'}

    initialized = False

    allow_snippet_cache = False

    def __init__(self, metadata: dict, panoply: Panoply = None):
        """
        Initialize a new PanosSkillet class.

        :param metadata: loaded dict from the Skillet YAML file

        :param panoply: optional panoply object. This can be passed in if the outer application scope has
        already been in contact with the device for things like checking auth, etc. If not passed in,
        you can invoke it in 'online' mode by passing in 'panos_username', 'panos_password' and 'panos_hostname' in the
        context. Otherwise, 'offline' mode requires a 'config' to be passed in via the context.
        """
        if panoply is not None:
            self.panoply = panoply
        super().__init__(metadata)

    def initialize_context(self, initial_context: dict) -> dict:
        """
        In this panos case, we want to stash the current configuration of the panos device in question in the
        context, check for online mode, offline mode, or an existing panoply object

        :param initial_context: dict to use to initialize the context
        :return: context with additional initialized items
        """

        # if the panoply object was not passed in via __init__, then check for online vs offline mode here
        # which set of fields we find in the context will determine online vs offline mode
        online_required_fields = {'panos_hostname', 'panos_username', 'panos_password'}

        # deprecated legacy fields
        legacy_required_fields = {'TARGET_IP', 'TARGET_USERNAME', 'TARGET_PASSWORD'}

        # simplified version
        provider_required_fields = {'ip_address', 'username', 'password'}

        # also allow api_key auth as well
        api_key_required_fields = {'ip_address', 'api_key'}

        # support for offline mode requires at least the 'config' variable to be present
        offline_required_fields = {'config'}

        context = super().initialize_context(initial_context)

        if self.panoply is None:
            if not online_required_fields.issubset(initial_context) \
                    and not offline_required_fields.issubset(initial_context) \
                    and not legacy_required_fields.issubset(initial_context) \
                    and not provider_required_fields.issubset(initial_context) \
                    and not api_key_required_fields.issubset(initial_context):
                raise SkilletValidationException('Required fields for panos skillet not found in context!')

            if online_required_fields.issubset(initial_context):
                hostname = initial_context.get('panos_hostname', None)
                username = initial_context.get('panos_username', None)
                password = initial_context.get('panos_password', None)
                port = initial_context.get('panos_port', '443')

                self.panoply = self.__init_panoply(hostname, username, password, port)

                context['config'] = self.panoply.get_configuration()

            elif legacy_required_fields.issubset(initial_context):
                hostname = initial_context.get('TARGET_IP', None)
                username = initial_context.get('TARGET_USERNAME', None)
                password = initial_context.get('TARGET_PASSWORD', None)
                port = initial_context.get('TARGET_PORT', '443')

                self.panoply = self.__init_panoply(hostname, username, password, port)

                context['config'] = self.panoply.get_configuration()

            elif provider_required_fields.issubset(initial_context):
                hostname = initial_context['ip_address']
                username = initial_context['username']
                password = initial_context['password']
                port = initial_context.get('port', 443)

                self.panoply = self.__init_panoply(hostname, username, password, port)

                context['config'] = self.panoply.get_configuration()

            elif api_key_required_fields.issubset(initial_context):
                hostname = initial_context['hostname']
                api_key = initial_context['api_key']

                # port may or may not be defined here
                port = initial_context.get('port', 443)

                self.panoply = self.__init_panoply(hostname=hostname, api_key=api_key, port=port)

                context['config'] = self.panoply.get_configuration()

            else:
                logger.info(f'offline mode detected for {__name__}')
                # init panoply in offline mode
                self.panoply = self.__init_panoply()

        else:
            # we were passed in a panoply object already, check if we are connected and grab the configuration if so
            if self.panoply.connected:
                context['config'] = self.panoply.get_configuration()

            else:
                raise SkilletLoaderException('Could not get configuration! Not connected to PAN-OS Device')

        self.initialized = True
        return context

    @staticmethod
    def __init_panoply(hostname: Optional[str] = None,
                       username: Optional[str] = None,
                       password: Optional[str] = None,
                       port: Optional[int] = 443,
                       api_key: Optional[str] = None):

        if hostname is None or username is None:
            # allow offline mode if these items are not passed in
            return skilletlib.panoply.EphemeralPanos()
        else:
            # otherwise, throw an exception when the device isn't ready
            return skilletlib.panoply.Panos(hostname=hostname,
                                            api_username=username,
                                            api_password=password,
                                            api_port=port,
                                            api_key=api_key)

    def get_snippets(self) -> List[PanosSnippet]:
        """
        Perform Panos Skillet specific tasks while loading each snippet

        :return: a List of PanosSnippets
        """
        if hasattr(self, 'snippets'):
            if self.initialized and self.allow_snippet_cache:
                return self.snippets

        snippet_list = list()

        for snippet_def in self.snippet_stack:

            if 'cmd' not in snippet_def or snippet_def['cmd'] == 'set':
                if 'element' not in snippet_def or snippet_def['element'] == '':
                    snippet_def['element'] = self.load_template(snippet_def['file'])

            snippet = PanosSnippet(snippet_def, self.panoply)
            snippet_list.append(snippet)

        if self.initialized:
            self.allow_snippet_cache = True

        return snippet_list

    @staticmethod
    def load_element(snippet_def: dict, snippet_path: Path) -> dict:
        """
        This method will load the snippet file found on disk into the 'element' attribute if the element
        is not already populated. This allows snippets to be 'all-in-one' i.e. there is no requirement for the snippets
        to be split into separate files. The meta-cnc.yaml file can contain all the snippets 'inline' in the 'element'
        attribute if desired.
        An example snippet def:

          - name: template
            xpath: /config/devices/entry[@name='localhost.localdomain']/template
            file: ../snippets/template.xml

        :param snippet_def: the loaded snippet definition from the .meta-cnc.yaml file. Each snippet object in the
        'snippets' stanza is a snippet_def and is passed in here
        :param snippet_path: the path on the filesystem where this skillet is located. This is used to resolve
        relative paths for each snippet. This allows snippet file re-use across skillets.
        :return: snippet_def with the element populated with the resolved and loaded snippet file contents
        """

        if 'element' not in snippet_def or snippet_def['element'] == '':

            if 'file' not in snippet_def:
                raise SkilletLoaderException(
                    'YAMLError: Could not parse metadata file for snippet %s' % snippet_def['name'])

            snippet_file = snippet_path.joinpath(snippet_def['file']).resolve()

            if snippet_file.exists():
                with snippet_file.open() as sf:
                    snippet_def['element'] = sf.read()

            else:
                # raise SkilletLoaderException('Could not load "file" attribute!')
                logger.error(f'Could not load the referenced file for {snippet_def["name"]}')

        else:
            # we have an element directly defined
            snippet_def['element'] = html.unescape(snippet_def['element'])

        return snippet_def

    def get_results(self) -> dict:
        """
        PanosSkillet will return a dict containing three keys:
        result, changed, and snippets. If any snippet failed, the result will be 'failure' otherwise 'success'
        If any successful snippet may have caused a change to the device, the 'changed' attribute will be 'True'

        A skillet that contains only the following snippet, will generate the output below:

        .. code-block:: yaml

              - name: check_hostname_again
                cmd: op
                cmd_str: <show><system><info/></system></show>
                outputs:
                  - name: url-db
                    capture_pattern: ./url-db
                  - name: pa-version
                    capture_pattern: ./plugin_versions/entry[@name="cloud_services"]/@version


        .. code-block:: json

            {
              'snippets': {
                'check_hostname_again': {
                  'results': 'success',
                  'changed': True
                }
              },
              'outputs': {
                'url-db': 'paloaltonetworks',
                'pa-version': '1.5.0'
              },
              'result': 'success',
              'changed': True
            }


        :return: dict containing default outputs plus the overall result and changed flag
        """

        results = super()._get_snippet_results()
        results['outputs'] = self.captured_outputs

        skillet_result = 'success'

        changed = False

        snippets = results.get('snippets', {})

        for k, v in snippets.items():
            if isinstance(v, dict) and 'results' in v:

                if v.get('results', 'failure') != 'success':
                    skillet_result = 'failure'
                else:
                    if v.get('changed', False):
                        changed = True

        results['result'] = skillet_result
        results['changed'] = changed

        return self._parse_output_template(results)



from .base import Snippet


class Python3Snippet(Snippet):
    """
    Basic Python3 Snippet Type
    """
    required_metadata = {'name', 'file'}

    def execute(self, context):
        """
        This is a stub function and is NOT currently implemented in skilletlib.

        :param context: The context used for python3 script execution
        :return: script output and tuple of success / failure
        """
        if 'python3_output' in context:
            # this is admittedly a hack to allow python3 skillets to be executed using other means in other tools
            # but still use output capturing, output_templates, etc from skilletlib. This is def a FIXME item
            return context['python3_output'], 'success'

        return '', 'success'



# from .panos import PanosSkillet
# from .docker import DockerSkillet
# from .pan_validation import PanValidationSkillet
# from .python3 import Python3Skillet
# from .rest import RestSkillet
# from .template import TemplateSkillet
# from .workflow import WorkflowSkillet



# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Adam Baumeister, Nathan Embery

import copy
import html
import logging
import os
import sys
import time
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Generator
from typing import List

import yaml
from yaml.scanner import ScannerError

from skilletlib.exceptions import SkilletLoaderException
from skilletlib.exceptions import SkilletValidationException
from skilletlib.snippet.base import Snippet
from skilletlib.snippet.template import SimpleTemplateSnippet

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

if not len(logger.handlers):
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)


class Skillet(ABC):
    # each skillet type can override this and set what metadata attributes are required
    snippet_required_metadata = {'name'}

    # optional metadata that can be present on each snippet
    snippet_optional_metadata = dict()

    def __init__(self, s: dict):
        """
        Initialize the base skillet type

        :param s: loaded dictionary from the .meta-cnc.yaml file
        """

        self.skillet_dict = self.__normalize_skillet_dict(s)
        self.name = self.skillet_dict['name']
        self.label = self.skillet_dict['label']
        self.description = self.skillet_dict['description']
        self.snippet_stack = self.skillet_dict['snippets']
        self.type = self.skillet_dict['type']
        self.supported_versions = 'not implemented'
        self.variables = self.__initialize_variables(s['variables'])
        # path is needed only when snippets are held in a relative file path
        self.path = self.skillet_dict.get('snippet_path', '')
        self.labels = self.skillet_dict['labels']
        self.collections = self.skillet_dict['labels'].get('collection', list())
        self.context = dict()
        self.captured_outputs = dict()
        self.snippet_outputs = dict()

        # ensure all values are set appropriately in the snippet definition
        self.__validate_snippet_metadata()

        # initialize our snippets
        self.snippets = self.get_snippets()

        # update our list of declared variables
        self.declared_variables = self.get_declared_variables()

        debug = os.environ.get('SKILLET_DEBUG', False)

        if debug:
            logger.setLevel(logging.DEBUG)
            logger.debug('Debugging output enabled')

    @abstractmethod
    def get_snippets(self) -> List[Snippet]:
        """
        Each skillet determines how it's snippets are to be loaded and initialized. Each Skillet type must
        implement this method.

        :return: List of Snippets for this Skillet Class
        """
        snippet_list = list()
        for snippet_def in self.snippet_stack:
            snippet = Snippet(snippet_def)
            snippet_list.append(snippet)

        return snippet_list

    def load_template(self, template_path: str) -> str:
        """
        Utility method to load a template file and return the contents as str

        :param template_path: relative path to the template to load
        :return: str contents
        """
        skillet_path = Path(self.path)
        template_file = skillet_path.joinpath(template_path).resolve()

        if template_file.exists():
            with template_file.open() as sf:
                return html.unescape(sf.read())

        else:
            # Add the snippet name here as well to allow for more context
            # fix for https://gitlab.com/panw-gse/as/panhandler/-/issues/19
            logger.error(f'Snippet: {self.name} has file attribute that does not exist')
            logger.error(f'Snippet file path is: {template_path}')
            raise SkilletLoaderException(f'Snippet: {self.name} - Could not resolve template path!')

    def update_context(self, d: dict) -> dict:
        """
        Take the input dict d and update the skillet context. I.e. any variables passed in via environment variables
        will be used to update the context stored on this skillet.

        :param d: dictionary of key value pairs. Any keys that match 'variable' keys will be used to update the context
        :return: updated context stored on this skillet
        """
        for var in self.variables:
            if var['name'] in d:
                self.context[var['name']] = d[var['name']]
            else:
                self.context[var['name']] = var['default']

        return self.context

    def initialize_context(self, initial_context: dict) -> dict:
        """
        Child classes can override this to provide any initialization information in the context. For example, 'panos'
        skillets use this to set up and initialize a Panos device object

        :param initial_context: Initial Context from user input, environment vars, etc
        :return: updated context with initial context items plus any initialization items
        """
        self.context.update(initial_context)
        self.update_context(initial_context)
        return self.context

    @staticmethod
    def __initialize_variables(vars_dict: dict) -> dict:
        """
        Ensure the proper default values are configured for each type of variable that may be present in the skillet

        :param vars_dict: Skillet 'variables' stanza
        :return: variables stanza with default values correctly parsed
        """

        for variable in vars_dict:
            default = variable.get('default', '')
            type_hint = variable.get('type_hint', 'text')
            if type_hint == "dropdown" and "dd_list" in variable:
                for item in variable.get('dd_list', []):
                    if 'key' in item and 'value' in item:
                        if default == item['key'] and default != item['value']:
                            # user set the key as the default and not the value, just fix it for them here
                            variable['default'] = item['value']
            elif type_hint == "radio" and "rad_list" in variable:
                rad_list = variable['rad_list']
                for item in rad_list:
                    if 'key' in item and 'value' in item:
                        if default == item['key'] and default != item['value']:
                            variable['default'] = item['value']
            elif type_hint == "checkbox" and "cbx_list" in variable:
                cbx_list = variable['cbx_list']
                for item in cbx_list:
                    if 'key' in item and 'value' in item:
                        if default == item['key'] and default != item['value']:
                            variable['default'] = item['value']

        return vars_dict

    def cleanup(self):
        pass

    def execute_async(self, initial_context: dict) -> Generator:
        """
        Returns a generator that can be used to iterate over the output as it's generated
        from each snippet. The calling application should call 'get_results' once the execute is complete

        :param initial_context: context of key values pairs to use for the execution. By default this is all the
        variables defined in the skillet file with their default values. Updates from user input, the environment, etc
        will override these default values via the 'update_context' method.
        :return: generator[str]
        """
        try:
            context = self.initialize_context(initial_context)
            logger.debug(f'Executing Async Skillet: {self.name}')

            for snippet in self.get_snippets():
                try:
                    # render anything that looks like a jinja template in the snippet metadata
                    # mostly useful for xpaths in the panos case
                    snippet.render_metadata(context)
                    # check the 'when' conditional against variables currently held in the context

                    if snippet.should_execute(context):
                        (output, status) = snippet.execute(context)
                        logger.debug(f'{snippet.name} - status: {status}')

                        if output:
                            logger.debug(f'{snippet.name} - output: {output}')

                        full_output = ''
                        while status == 'running':
                            # logger.info('Snippet still running...')
                            time.sleep(5)
                            (partial_output, status) = snippet.get_output()

                            full_output += partial_output

                            yield partial_output
                            output = full_output

                        # capture all outputs
                        snippet_outputs = snippet.get_default_output(output, status)
                        captured_outputs = snippet.capture_outputs(output, status)

                        if captured_outputs:
                            logger.debug(f'{snippet.name} - captured_outputs: {captured_outputs}')

                        self.snippet_outputs.update(snippet_outputs)
                        self.captured_outputs.update(captured_outputs)

                        context.update(snippet_outputs)
                        context.update(captured_outputs)

                except SkilletLoaderException as sle:
                    logger.error(f'Caught Exception during execution: {sle}')
                    snippet_outputs = snippet.get_default_output(str(sle), 'error')
                    logger.error(snippet_outputs)
                    self.snippet_outputs.update(snippet_outputs)

                except Exception as e:
                    logger.error(f'Exception caught: {e}')
                    snippet_outputs = snippet.get_default_output(str(e), 'error')
                    self.snippet_outputs.update(snippet_outputs)

        finally:
            self.cleanup()

        return None

    def execute(self, initial_context: dict) -> dict:
        """
        The heart of the Skillet class. This method executes the skillet by iterating over all the skillets returned
        from the 'get_skillets' method. Each one is checked if it should be executed if a 'when' conditional attribute
        is found, and if so, is executed using the snippet execute method.

        :param initial_context: context of key values pairs to use for the execution. By default this is all the
        variables defined in the skillet file with their default values. Updates from user input, the environment, etc
        will override these default values via the 'update_context' method.
        :return: a dict containing the updated context containing the output of each of the snippets
        """
        try:
            context = self.initialize_context(initial_context)
            logger.debug(f'Executing Skillet: {self.name}')

            for snippet in self.get_snippets():
                try:
                    # render anything that looks like a jinja template in the snippet metadata
                    # mostly useful for xpaths in the panos case
                    snippet.render_metadata(context)
                    # check the 'when' conditional against variables currently held in the context

                    if snippet.should_execute(context):
                        (output, status) = snippet.execute(context)
                        logger.debug(f'{snippet.name} - status: {status}')

                        if output:
                            logger.debug(f'{snippet.name} - output: {output}')

                        running_counter = 0

                        while status == 'running':
                            logger.info('Snippet still running...')
                            time.sleep(5)
                            (output, status) = snippet.get_output()
                            running_counter += 1

                            if running_counter > 60:
                                raise SkilletLoaderException('Snippet took too long to execute!')

                        # capture all outputs
                        snippet_outputs = snippet.get_default_output(output, status)
                        captured_outputs = snippet.capture_outputs(output, status)

                        if captured_outputs:
                            logger.debug(f'{snippet.name} - captured_outputs: {captured_outputs}')

                        self.snippet_outputs.update(snippet_outputs)
                        self.captured_outputs.update(captured_outputs)

                        context.update(snippet_outputs)
                        context.update(captured_outputs)

                except SkilletLoaderException as sle:
                    logger.error(f'Caught Exception during execution: {sle}')
                    snippet_outputs = snippet.get_default_output(str(sle), 'error')
                    logger.error(snippet_outputs)
                    self.snippet_outputs.update(snippet_outputs)

                except Exception as e:
                    logger.error(f'Exception caught: {e}')
                    snippet_outputs = snippet.get_default_output(str(e), 'error')
                    self.snippet_outputs.update(snippet_outputs)

        finally:
            self.cleanup()

        return self.get_results()

    def get_results(self) -> dict:
        """
        Returns the results from the skillet execution. This must be called manually if using 'execute_async'. The
        returned dict will include a 'snippets' dictionary that contains a key for each snippet that was executed. Each
        snippet dictionary will contain the 'results' and 'raw' attributes.

        .. code-block:: json

            {
              'snippets': {
                'check_hostname_again': {
                  'results': 'success',
                  'changed': True
                }
              },
              'outputs': {
                'url-db': 'paloaltonetworks',
                'pa-version': '1.5.0'
              },
              'result': 'success',
              'changed': True
            }


        :return: dictionary of results from the Skillet execute or execute_async method
        """

        results = self._get_snippet_results()

        results['outputs'] = self.captured_outputs
        return self._parse_output_template(results)

    def _get_snippet_results(self) -> dict:
        """
        Iterate the snippets in the skillet and collect their combined results into the 'results' dict

        :return: dict with 'snippets' key
        """
        results = dict()
        results['snippets'] = dict()

        for s in self.snippet_stack:
            snippet_name = s.get('name', '')

            if snippet_name in self.snippet_outputs:
                results['snippets'][snippet_name] = self.snippet_outputs[snippet_name]

        return results

    def _parse_output_template(self, results: dict) -> dict:
        """
        Check for the presence of the output_template label. If found, load and parse the template
        using the results. The parsed output will be placed in the 'output_template' key in the results

        :param results: results of the skillet execution
        :return: results plus the output_template value if found
        """
        try:
            if 'output_template' in self.labels:
                output_template = self.load_template(self.labels['output_template'])
                template_snippet = SimpleTemplateSnippet(output_template)
                # create context dict for template parsing
                # add all outputs as 'top-level' attributes
                context = results.get('outputs', {})
                # add other keys as top-level attributes as well...
                for k, v in results.items():
                    if k != 'outputs':
                        context[k] = v
                results['output_template'] = template_snippet.template(context)

        except SkilletLoaderException as sle:
            print(sle)

        return results

    def __validate_snippet_metadata(self) -> None:
        """
        Perform snippet metadata validation before we attempt to instantiate the snippet

        This will throw a SkilletLoaderException if a required attribute is not present in the metadata
        Will also set all optional metadata attributes with their default values

        :raises: SkilletLoaderException if a required field is not present
        :return: None
        """
        for s in self.snippet_stack:
            name = s.get('name', '')
            for r in self.snippet_required_metadata:
                if r not in s:
                    raise SkilletLoaderException(f'Invalid snippet metadata configuration: attribute: {r} '
                                                 f'is required for snippet: {name}')

            for k, v in self.snippet_optional_metadata.items():
                if k not in s:
                    s[k] = v

    def get_declared_variables(self) -> List[str]:
        """
        Return a list of all variables defined in all the snippets that are not defined as an output

        :return: list of variable names
        """

        # get list of output_vars from all snippets using double list comprehension
        output_vars = [o for s in self.get_snippets() for o in s.get_output_variables()]

        # get list of all variables defined in all snippets that are NOT in the output_vars
        dv = [x for s in self.get_snippets() for x in s.get_snippet_variables() if x not in output_vars]

        # convert to set and back to list to remove dups
        return list(set(dv))

    @staticmethod
    def __normalize_skillet_dict(skillet: dict) -> dict:

        if skillet is None:
            skillet = dict()

        if type(skillet) is not dict:
            skillet = dict()

        if 'name' not in skillet:
            skillet['name'] = 'Unknown Skillet'

        if 'label' not in skillet:
            skillet['label'] = 'Unknown Skillet'

        if 'type' not in skillet:
            skillet['type'] = 'template'

        if 'description' not in skillet:
            skillet['description'] = 'Unknown Skillet'

        if 'variables' not in skillet:
            skillet['variables'] = list()

        if 'snippets' not in skillet:
            skillet['snippets'] = list()

        if 'labels' not in skillet:
            skillet['labels'] = dict()

        if 'collection' not in skillet['labels']:
            skillet['labels']['collection'] = list()
            skillet['labels']['collection'].append('Kitchen Sink')

        return skillet

    def dump_yaml(self) -> str:
        """
        Convert this Skillet into a YAML formatted string

        :return: YAML formatted string
        """
        try:

            # remove non-essential non-portable items from the dict before dumping
            safe_skillet_dict = copy.deepcopy(self.skillet_dict)

            safe_skillet_dict.pop('snippet_path', None)
            safe_skillet_dict.pop('app_data', None)

            # source https://stackoverflow.com/a/45004775
            yaml.SafeDumper.org_represent_str = yaml.SafeDumper.represent_str

            def repr_str(dumper, data):
                if '\n' in data:
                    return dumper.represent_scalar(u'tag:yaml.org,2002:str', data, style='|')

                return dumper.org_represent_str(data)

            yaml.add_representer(str, repr_str, Dumper=yaml.SafeDumper)

            return yaml.safe_dump(safe_skillet_dict, indent=4)

        except (ScannerError, ValueError) as err:
            raise SkilletValidationException(f'Could not dump Skillet as YAML: {err}')



# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Adam Baumeister, Nathan Embery

from pathlib import Path
from typing import List

from skilletlib.snippet.pan_validation import PanValidationSnippet
from .panos import PanosSkillet


# from .base import Skillet


class PanValidationSkillet(PanosSkillet):
    snippet_list = list()

    def get_snippets(self) -> List[PanValidationSnippet]:

        if hasattr(self, 'snippets'):
            if self.initialized and self.allow_snippet_cache:
                return self.snippets

        snippet_path_str = self.skillet_dict.get('snippet_path', '')
        snippet_path = Path(snippet_path_str)
        snippet_list = list()
        for snippet_def in self.snippet_stack:
            if 'cmd' not in snippet_def:
                snippet_def['cmd'] = 'validate'
            elif snippet_def['cmd'] == 'validate_xml':
                snippet_def = self.load_element(snippet_def, snippet_path)
            elif snippet_def['cmd'] == 'set':
                snippet_def = self.load_element(snippet_def, snippet_path)

            if 'severity' not in snippet_def:
                snippet_def['severity'] = 'low'

            snippet = PanValidationSnippet(snippet_def, self.panoply)
            snippet_list.append(snippet)

        if self.initialized:
            self.allow_snippet_cache = True

        return snippet_list

    def get_results(self) -> dict:
        """
        Pan-validation skillets return a dictionary with a key for each test that was executed. Each value of those
        keys will be a dict containing the following keys:
            * results - whether the test conditional was true or false
            * label - human readable label of the test
            * severity - a string that may be set to indicate the severity of a test
            * documentation_link - an HTTP link where the user can get more information about this test
            * output_message - A rendered output message regarding the test results

        .. code-block:: json

                {
                    "update_schedule_configured": {
                        "results": true,
                        "label": "Ensure Update Schedules are Configured",
                        "severity": "low",
                        "documentation_link": "https://iron-skillet.readthedocs.io",
                        "test": "update_schedule_object is not none",
                        "output_message": "Snippet Validation Passed"
                    },
                }


        :return: dictionary with the aforementioned keys
        """
        # do not call super() as this subclasses panos and not base directly
        results = dict()
        results['snippets'] = dict()
        results['pan_validation'] = dict()
        context = self.context

        for s in self.get_snippets():
            snippet_name = s.name
            cmd = s.cmd
            # handle both validate and validate_xml here
            if snippet_name in context and 'validate' in cmd:
                if 'results' in context[snippet_name]:
                    result = context[snippet_name]['results']
                    label_template = context[snippet_name].get('label', '')
                    # attempt to render the label using supplied context
                    context[snippet_name]['label'] = s.render(label_template, context)
                    if not result:
                        fail_message = s.metadata.get('fail_message', 'Snippet Validation results were {{ result }}')
                        context[snippet_name]['output_message'] = s.render(fail_message, context)
                    elif result:
                        pass_message = s.metadata.get('pass_message', 'Snippet Validation results were {{ result }}')
                        context[snippet_name]['output_message'] = s.render(pass_message, context)
                    else:
                        context[snippet_name]['output_message'] = 'Unknown results from Snippet Validation'

                    results['snippets'][snippet_name] = result

                results['pan_validation'][snippet_name] = context[snippet_name]

        return self._parse_output_template(results)



# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Nathan Embery

import json
import logging
import re
import xml.etree.ElementTree as elementTree
from abc import ABC
from abc import abstractmethod
from base64 import urlsafe_b64encode
from typing import Tuple
from xml.etree.ElementTree import ParseError

import xmltodict
from jinja2 import BaseLoader
from jinja2 import Environment
from jinja2 import meta
from jinja2.exceptions import TemplateAssertionError
from jinja2.exceptions import UndefinedError
from jsonpath_ng import parse
from lxml import etree
from passlib.hash import md5_crypt

from jinja2_ansible_filters import AnsibleCoreFiltersExtension

from skilletlib.exceptions import SkilletLoaderException
from skilletlib.exceptions import SkilletValidationException

logger = logging.getLogger(__name__)


class Snippet(ABC):
    """
    BaseSnippet implements a basic Noop snippet
    """
    # set of required metadata, each snippet will define what attributes are required in the snippet definition
    # by default, we only require a 'name' attribute, but sub-classes will require more
    required_metadata = {'name'}

    # dict of optional metadata  and their default values. These values will be set on the snippet class but
    # will not throw an exception is they are not present
    optional_metadata = dict()

    # metadata fields that should be considered templates and rendered
    template_metadata = set()

    # set a default output type. this can be overridden for each SnippetType. This is used to determine the default
    # output handler to use for each snippet class. This can be set on a per snippet basis, but this allows a
    # short-cut on each
    output_type = 'xml'

    def __init__(self, metadata):

        # first validate all the required fields are present in the metadata (snippet definition)
        self.metadata = self.sanitize_metadata(metadata)
        # always have a default name, subclasses will set additional fields on the class
        self.name = self.metadata['name']
        # set up jinja environment and add any custom filters. Snippet sub-classes can override __add_filters
        # to append additional filters. See the PanosSnippet class for an example
        self.__init_env()

        # set all the required fields with their values from the snippet definition
        for k in self.required_metadata:
            setattr(self, k, self.metadata[k])

        # iterate the optional_metadata dict and set the default values
        # if they have not been set in the snippet metadata directly
        for k, v in self.optional_metadata.items():
            if v is None:
                continue
            if k in self.metadata:
                setattr(self, k, self.metadata[k])
            else:
                setattr(self, k, v)
                self.metadata[k] = v

        self.context = dict()

    def update_context(self, context: dict) -> dict:
        """
        This will update the snippet context with the passed in dict.
        This gets called inside of 'should_execute'

        :param context: dict of the outer context
        :return: newly updated context
        """
        self.context.update(context)
        return self.context

    @abstractmethod
    def execute(self, context: dict) -> Tuple[str, str]:
        """
        Execute this Snippet and return a tuple consisting on raw output and a string representing
        success, failure, or running.

        Each snippet sub class must override this method!

        :param context: context to use for variable interpolation
        :return: Tuple containing raw snippet output and string indicated success or failure
        """
        return '', 'success'

    def should_execute(self, context: dict) -> bool:
        """
        Evaluate 'when' conditionals and return a bool if this snippet should be executed

        :param context: jinja context containing previous outputs and user supplied variables
        :return: boolean
        """

        logger.debug(f'Checking snippet: {self.name}')

        # hook for pre-conditional checks
        context = self.update_context(context)

        # check if this snippet should be filtered out of execution based on the context object __filter_snippets
        if self.is_filtered(context):
            logger.debug('Skipping this snippet due to snippet inclusion rules')
            return False

        if 'when' not in self.metadata:
            # always execute when no when conditional is present
            logger.debug(f'No conditional present, proceeding with skillet: {self.name}')
            return True

        results = self.execute_conditional(self.metadata['when'], context)
        logger.debug(f'  Conditional Evaluation results: {results} ')
        return results

    def is_filtered(self, context) -> bool:
        """
        Determines if a snippet should be available for execution based on the presence of the `__filter_snippets`
        object in the context. Snippets can be filtered by the following:

        include_by_name: list of names to check. Only snippet names included in this list will be executed
        include_by_tag: list of tags to check. Only snippets with those tags will be executed
        include_by_regex: regular expression match. Only snippets whose name matches the regex will be executed

        any snippet that does not match any of the above rules will be filtered out. The rules are inclusive OR

        :param context: Snippet Context
        :return: bool
        """

        # by default, nothing is filtered out
        is_filtered = None

        if '__filter_snippets' in context:
            logger.debug('filtering snippet...')

            fc = context['__filter_snippets']
            if type(fc) is not dict:
                logger.warning('Snippet filter is malformed...')
                return False
            for filter_def in ('include_by_tag', 'include_by_name', 'include_by_regex', 'exclude_by_tag',
                               'exclude_by_name', 'exclude_by_regex'):
                if filter_def in fc:
                    if type(fc[filter_def]) is list:
                        for item in fc[filter_def]:
                            is_filtered = self.__consider_filter(filter_def, item)
                    elif type(fc[filter_def]) is str:
                        item = fc[filter_def]
                        is_filtered = self.__consider_filter(filter_def, item)

                # we have discovered this snippet should not be filtered, jump out now
                if is_filtered is not None:
                    return is_filtered

            # we have considered all rules, and still no determination, however, because we have some rules due to
            # presence of '__filter_snippets' any item that is not specifically included or excluded, should be
            # excluded

            if is_filtered is None:
                return True
        else:
            # there is not filter snippets config present, so include all snippets by default
            return False

    def __consider_filter(self, filter_def: str, item: str) -> (bool, None):
        """
        Rules that positively match are returned immediately, any snippet that does NOT specifically match any rule
        is excluded (True response == snippet is NOT executed)

        :param filter_def: type of filter to consider
        :param item: specific filter item
        :return: bool or None if not match
        """
        if filter_def == 'include_by_name':
            # this name matches, do NOT filter it out
            if self.name == item:
                return False

        elif filter_def == 'exclude_by_name':
            # name matches exclusion rule, filter it out
            if self.name == item:
                return True

        if filter_def == 'include_by_tag':
            if self.__has_tag(item):
                # snippet has this tag, do not filter out
                return False

        elif filter_def == 'exclude_by_tag':
            if self.__has_tag(item):
                # snippet has this tag, filter it out
                return True

        if '_by_regex' in filter_def:
            match = re.match(item, self.name)
            if 'include_by_regex' == filter_def:
                if match:
                    return False
            elif 'exclude_by_regex' == filter_def:
                if match:
                    return True

        # this snippet does not match any of the rules above, one way or the other, so filter it out of consideration
        return None

    def __has_tag(self, tag_to_check: str):
        if 'tag' in self.metadata and type(self.metadata['tag']) is list:
            for tag in self.metadata['tag']:
                if tag_to_check == tag:
                    return True

            return False

        elif 'tag' in self.metadata and type(self.metadata['tag']) is str:
            return self.metadata['tag'] == tag_to_check
        else:
            return False

    def execute_conditional(self, test: str, context: dict) -> bool:
        """
        Evaluate 'test' conditionals and return a bool

        :param test: string of the conditional to execute
        :param context: jinja context containing previous outputs and user supplied variables
        :return: boolean
        """
        try:
            test_str = '{{%- if {0} -%}} True {{%- else -%}} False {{%- endif -%}}'.format(test)
            test_template = self._env.from_string(test_str)
            results = test_template.render(context)
            if str(results).strip() == 'True':
                return True
            else:
                return False
        except UndefinedError as ude:
            logger.error(ude)
            # always return false on error condition
            return False
        except TypeError as te:
            logger.error(te)
            return False
        except TemplateAssertionError as tae:
            logger.error(tae)
            raise SkilletValidationException('Malformed Jinja expression!')
        except Exception as e:
            # catch-all - always return false on other error conditions
            logger.error(e)
            logger.error(type(e))
            return False

    def get_output(self) -> Tuple[str, str]:
        """
        get_output can be used when a snippet executes async and cannot or will not return output right away
        snippets that operate async must override this method

        :return: Tuple containing the skillet output as a str and a str indicating success of failure
        """

        return '', 'success'

    def get_default_output(self, results: str, status: str) -> dict:
        """
        each snippet type can override this method to provide it's own default output. This is used
        when there are no variables defined to be captured

        :param results: raw output from snippet execution
        :param status: status of the snippet.execute method
        :return: dict of default outputs
        """

        r = {
            self.name: {
                'results': status,
                'raw': results
            }
        }
        return r

    def capture_outputs(self, results: str, status: str) -> dict:
        """
        All snippet output or portions of snippet output can be captured and saved on the context as a new variable

        :param results: the raw output from the snippet execution
        :param status: status of the snippet.execute method
        :return: a dictionary containing all captured variables
        """

        # always capture the default output
        # captured_outputs = self.get_default_output(results, status)
        captured_outputs = dict()
        output_type = self.metadata.get('output_type', self.output_type)

        # check if this snippet type wants to handle it's own outputs
        if hasattr(self, f'handle_output_type_{output_type}'):
            func = getattr(self, f'handle_output_type_{output_type}')
            return func(results)

        # otherwise, check all the normal types here
        if 'outputs' not in self.metadata:
            return captured_outputs

        for output in self.metadata['outputs']:

            outputs = dict()

            if 'name' not in output:
                continue

            if not results:
                outputs[output['name']] = ''
                continue

            if 'capture_variable' in output:
                outputs[output['name']] = self.render(output['capture_variable'], self.context)

            elif 'capture_expression' in output:
                expression = self._env.compile_expression(output['capture_expression'])
                value = expression(self.context)
                outputs[output['name']] = value

            else:
                # allow jinja syntax in capture_pattern, capture_value, capture_object etc
                output = self.__render_output_metadata(output, self.context)

                if output_type == 'xml':
                    outputs = self.__handle_xml_outputs(output, results)
                elif output_type == 'manual':
                    outputs = self.__handle_manual_outputs(output, results)
                elif output_type == 'text':
                    outputs = self.__handle_text_outputs(output, results)
                elif output_type == 'json':
                    outputs = self.__handle_json_outputs(output, results)
            # elif output_type == 'base64':
            #     outputs = self._handle_base64_outputs(results)

            # elif output_type == 'manual':
            #     outputs = self._handle_manual_outputs(results)
            # elif output_type == 'text':
            #     outputs = self.__handle_text_outputs(results)
            # # sub classes can handle their own output types
            # # see panos/__handle_validation for example
            # elif hasattr(self, f'handle_output_type_{output_type}'):
            #     func = getattr(self, f'handle_output_type_{output_type}')
            #     outputs = func(results)
            captured_outputs.update(outputs)
            self.context.update(outputs)

        return captured_outputs

    def __render_output_metadata(self, output: dict, context: dict) -> dict:
        # fix for #78 allow filter_items to be rendered
        keys = ('capture_value', 'capture_pattern', 'capture_object', 'capture_list', 'filter_items')
        for k in keys:
            if k in output:
                output[k] = self.render(output[k], context)

        return output

    def __filter_outputs(self, output_definition: dict, output: (str, dict, list), local_context: dict) -> (list, None):
        """
        Filter OUT items that do not pass the test

        :param output_definition: the output definition from the skillet
        :param output: the captured object to test
        :param local_context: local context for the jinja expression based tests
        :return: a list of the items that passed the test, or all items if there is not test defined
        """
        if 'filter_items' not in output_definition:
            return output

        # grab the test string to evaluate
        test_str = output_definition['filter_items']

        # keep a new list of all the items that have matched the test
        filtered_items = list()

        if isinstance(output, list):
            for item in output:
                local_context['item'] = item
                results = self.execute_conditional(test_str, local_context)
                if results:
                    filtered_items.append(item)

            # if len(filtered_items) == 0:
            #     output = None
            # elif len(filtered_items) == 1:
            #     output = filtered_items[0]
            # else:
            #     output = filtered_items
            return filtered_items

        elif isinstance(output, str) or isinstance(output, dict):
            local_context['item'] = output
            results = self.execute_conditional(test_str, local_context)
            if results:
                filtered_items.append(output)

        return filtered_items

    def render(self, template_str: str, context: (dict, None)) -> str:
        """
        Convenience method to quickly render a template_str using the provided context

        :param template_str: jinja2 template to render
        :param context: context to pass to the jinja2 environment
        :return: rendered string
        """
        if context is None:
            context = self.context
        t = self._env.from_string(template_str)
        return t.render(context)

    def get_variables_from_template(self, template_str: str) -> list:
        """
        Returns a list of jinja2 variable found in the template

        :param template_str: jinja2 template
        :return: list of variables declared in the template
        """

        parsed_template_str = self._env.parse(template_str)
        return meta.find_undeclared_variables(parsed_template_str)

    def get_output_variables(self) -> list:
        """
        Returns a list of all output variables. This is used to determine if a snippet variable should be considered
        undeclared.

        :return: list of str representing output variable names
        """

        return [x['name'] for x in self.metadata.get('outputs', dict()) if 'name' in x]

    def get_snippet_variables(self) -> list:
        """
        Returns a list of variables defined in this snippet that are NOT defined as outputs

        :return: list of str representing variables found in the jinja templates
        """

        variables = list()
        for i in self.template_metadata:
            if i in self.metadata:
                found_vars = self.get_variables_from_template(self.metadata[i])
                for f in found_vars:
                    if f not in variables:
                        variables.append(f)

        return variables

    def sanitize_metadata(self, metadata: dict) -> dict:
        """
        method to sanitize metadata. Each snippet type can override this provide extra logic over and above
        just checking the required and optional fields

        :param metadata: snippet metadata
        :return: sanitized snippet metadata
        """
        return metadata

    def render_metadata(self, context: dict) -> dict:
        """
        Each snippet sub class can override this method to perform jinja variable interpolation on various items
        in it's snippet definition. For example, the PanosSnippet will check the 'xpath' attribute and perform
        the required interpolation.

        This handles regular strings, lists, and dictionaries such as:

        in snippet class
        template_metadata = {'render_me', 'render_all', 'render_list'}

        in metadata
        snippets:
         - name: render_me_snippet
           render_me: render_{{ this }}
         - name: render_all_snippet
           render_all:
             a_key: some_{{ value }}
             another_key: some_other_{{ value }}
         - name: render_list_snippet
           render_list:
             - here_is_a_{{ value }}
             - another_{{ value }}

        :param context: context from environment
        :return: metadata with jinja rendered variables
        """
        self.context.update(context)

        # render all template metadata fields
        for key_name in self.template_metadata:

            if key_name in self.metadata:
                key = self.metadata[key_name]

                if isinstance(key, str):
                    rendered_str = self.render(key, context)
                    self.metadata[key_name] = rendered_str

                elif isinstance(key, dict):
                    rendered_dict = dict()
                    for k, v in key.items():
                        rendered_dict[k] = self.render(v, context)

                    self.metadata[key_name] = rendered_dict

                elif isinstance(key, list):
                    rendered_list = list()
                    for v in key:
                        rendered_list.append(self.render(v, context))

                    self.metadata[key_name] = rendered_list

        return self.metadata

    # define functions for custom jinja filters
    @staticmethod
    def __md5_hash(txt: str) -> str:
        """
        Returns the MD5 Hashed secret for use as a password hash in the PAN-OS configuration

        :param txt: text to be hashed
        :return: password hash of the string with salt and configuration information. Suitable to place in the phash field
        in the configurations
        """

        return md5_crypt.hash(txt)

    def __init_env(self) -> None:
        """
        init the jinja2 environment and add any required filters

        :return: Jinja2 environment object
        """
        self._env = Environment(loader=BaseLoader, extensions=[AnsibleCoreFiltersExtension])
        self._env.filters["md5_hash"] = self.__md5_hash
        self.add_filters()

    def add_filters(self) -> None:
        """
        Each snippet sub-class can add additional filters. See the PanosSnippet for examples

        :return: None
        """
        pass

    def __handle_text_outputs(self, output_definition: dict, results: str) -> dict:
        """
        Parse the results string as a text blob into a single variable.

        - name: system_info
          path: /api/?type=op&cmd=<show><system><info></info></system></show>&key={{ api_key }}
          output_type: text
          outputs:
            - name: system_info_as_xml

        :param results: results string from the action
        :return: dict of outputs, in this case a single entry
        """
        outputs = dict()
        output_name = output_definition.get('name', self.name)
        outputs[output_name] = ''

        # enhancement for https://gitlab.com/panw-gse/as/skilletlib/-/issues/86
        if 'capture_value' in output_definition:
            # allow capture_value to be equivalent to capture_pattern
            output_definition['capture_pattern'] = output_definition['capture_value']

        if 'capture_object' in output_definition:
            # allow capture_object to be equivalent to capture_list
            output_definition['capture_list'] = output_definition['capture_object']

        if 'capture_pattern' in output_definition:
            # this is a regex pattern we should use for a match
            pattern = re.compile(output_definition['capture_pattern'])
            matches = pattern.findall(results)
            if matches:
                # capture pattern should only return the first match
                outputs[output_name] = matches[0]

        elif 'capture_list' in output_definition:
            # this is a regex pattern we should use for a match
            pattern = re.compile(output_definition['capture_list'])
            matches = pattern.findall(results)
            if matches:
                # capture list should only the full list of matches
                outputs[output_name] = matches
            else:
                # no matches should return an empty list
                outputs[output_name] = list()

        else:
            output_name = output_definition.get('name', self.name)
            outputs[output_name] = results

        return outputs

    def __handle_xml_outputs(self, output_definition: dict, results: str) -> dict:
        """
        Parse the results string as an XML document
        Example .meta-cnc snippets section:
        snippets:

          - name: system_info
            path: /api/?type=op&cmd=<show><system><info></info></system></show>&key={{ api_key }}
            output_type: xml
            outputs:
              - name: hostname
                capture_value: result/system/hostname
              - name: uptime
                capture_value: result/system/uptime
              - name: sw_version
                capture_value: result/system/sw-version


        :param results: string as returned from some action, to be parsed as XML document
        :return: dict containing all outputs found from the capture pattern in each output
        """

        captured_output = dict()

        def unique_tag_list(elements: list) -> bool:
            tag_list = list()
            for el in elements:
                # some xpath queries can return a list of str
                if isinstance(el, str):
                    return False

                if el.tag not in tag_list:
                    tag_list.append(el.tag)

            if len(tag_list) == 1:
                # all tags in this list are the same
                return False
            else:
                # there are unique tags in this list
                return True

        try:
            xml_doc = etree.XML(results)

            # xml_doc = elementTree.fromstring(results)
            # allow jinja syntax in capture_pattern, capture_value, capture_object etc

            local_context = self.context.copy()
            output = self.__render_output_metadata(output_definition, local_context)

            var_name = output['name']
            if 'capture_pattern' in output or 'capture_value' in output:

                if 'capture_value' in output:
                    capture_pattern = output['capture_value']
                else:
                    capture_pattern = output['capture_pattern']

                # by default we will attempt to return the text of the found element
                return_type = 'text'
                entries = xml_doc.xpath(capture_pattern)
                logger.debug(f'found entries: {entries}')
                if len(entries) == 0:
                    captured_output[var_name] = ''
                elif len(entries) == 1:
                    entry = entries.pop()
                    if isinstance(entry, str):
                        captured_output[var_name] = str(entry)
                    else:
                        if len(entry) == 0:
                            # this tag has no children, so try to grab the text
                            if return_type == 'text':
                                captured_output[var_name] = str(entry.text).strip()
                            else:
                                captured_output[var_name] = entry.tag
                        else:
                            # we have 1 Element returned, so the user has a fairly specific xpath
                            # however, this element has children itself, so we can't return a text value
                            # just return the tag name of this element only
                            captured_output[var_name] = entry.tag
                else:
                    # we have a list of elements returned from the users xpath query
                    capture_list = list()
                    # are there unique tags in this list? or is this a list of the same tag names?
                    if unique_tag_list(entries):
                        return_type = 'tag'
                    for entry in entries:
                        if isinstance(entry, str):
                            capture_list.append(entry)
                        else:
                            if len(entry) == 0:
                                if return_type == 'text':
                                    if entry.text is not None:
                                        capture_list.append(entry.text.strip())
                                    else:
                                        # If there is no text, then try to grab a sensible attribute
                                        # if you need more control than this, then you should first
                                        # capture_object to convert to a python object then use a jinja filter
                                        # to get what you need
                                        if 'value' in entry.attrib:
                                            capture_list.append(entry.attrib.get('value', ''))
                                        elif 'name' in entry.attrib:
                                            capture_list.append(entry.attrib.get('name', ''))
                                        else:
                                            capture_list.append(json.dumps(dict(entry.attrib)))
                                else:
                                    capture_list.append(entry.tag)
                            else:
                                capture_list.append(entry.tag)

                    captured_output[var_name] = capture_list

            elif 'capture_object' in output:
                capture_pattern = output['capture_object']
                entries = xml_doc.xpath(capture_pattern)

                if len(entries) == 0:
                    captured_output[var_name] = None
                elif len(entries) == 1:
                    captured_output[var_name] = xmltodict.parse(elementTree.tostring(entries.pop()))
                else:
                    capture_list = list()
                    for entry in entries:
                        capture_list.append(xmltodict.parse(elementTree.tostring(entry)))
                    captured_output[var_name] = capture_list

            elif 'capture_list' in output:
                capture_pattern = output['capture_list']
                entries = xml_doc.xpath(capture_pattern)

                capture_list = list()
                for entry in entries:
                    if isinstance(entry, str):
                        capture_list.append(entry)
                    else:
                        capture_list.append(xmltodict.parse(elementTree.tostring(entry)))

                captured_output[var_name] = capture_list

            # filter selected items here
            captured_output[var_name] = self.__filter_outputs(output, captured_output[var_name], local_context)

        except ParseError:
            logger.error('Could not parse XML document in output_utils')
            # just return blank captured_outputs here
            raise SkilletLoaderException(f'Could not parse output as XML in {self.name}')

        return captured_output

    def _handle_base64_outputs(self, results: str) -> dict:
        """
        Parses results and returns a dict containing base64 encoded values

        :param results: string as returned from some action, to be encoded as base64
        :return: dict containing all outputs found from the capture pattern in each output
        """

        outputs = dict()

        snippet_name = 'unknown'
        if 'name' in self.metadata:
            snippet_name = self.metadata['name']

        try:
            if 'outputs' not in self.metadata:
                logger.info(f'No output defined in this snippet {snippet_name}')
                return outputs

            for output in self.metadata['outputs']:
                if 'name' not in output:
                    continue

                results_as_bytes = bytes(results, 'utf-8')
                encoded_results = urlsafe_b64encode(results_as_bytes)
                var_name = output['name']
                outputs[var_name] = encoded_results.decode('utf-8')

        except TypeError:
            raise SkilletLoaderException(f'Could not base64 encode results {snippet_name}')

        return outputs

    def __handle_json_outputs(self, output_definition: dict, results: str) -> dict:
        """
        Parses results using jsonpath_ng query syntax
        output_type: json
        outputs:
          - name: salt_auth_token
            capture_object: '$.return[0].token'

        See here for more jsonpath examples: https://github.com/h2non/jsonpath-ng

        :param results: string as returned from some action, to be parsed as JSON
        :return: dict containing all outputs found from the capture pattern in each output
        """
        captured_output = dict()

        local_context = self.context.copy()
        output = self.__render_output_metadata(output_definition, local_context)

        try:
            for i in ('capture_pattern', 'capture_value', 'capture_object'):
                if i in output:
                    capture_pattern = output[i]
                else:
                    continue

                if 'name' not in output:
                    continue

                # some Skillet types may return us json already, check if results are actually a str like object
                # before trying to convert
                if type(results) is not str or type(results) is not bytes or type(results) is not bytearray:
                    json_object = results
                else:
                    json_object = json.loads(results)

                var_name = output['name']

                # short cut for just getting all the results
                if capture_pattern == '$' or capture_pattern == '.':
                    captured_output[var_name] = json_object
                    continue

                jsonpath_expr = parse(capture_pattern)
                result = jsonpath_expr.find(json_object)
                if len(result) == 1:
                    captured_output[var_name] = str(result[0].value)
                else:
                    # FR #81 - add ability to capture from a list
                    capture_list = list()
                    for r in result:
                        capture_list.append(r.value)

                    captured_output[var_name] = capture_list

        except ValueError as ve:
            logger.error('Caught error converting results to json')
            captured_output['fail_message'] = str(ve)
        except Exception as e:
            logger.error('Unknown exception here!')
            logger.error(e)
            captured_output['fail_message'] = str(e)

        return captured_output

    def __handle_manual_outputs(self, output_definition: dict, results: str) -> dict:
        """
        Manually set a value in the context, this could be useful with 'when' conditionals

        :param results: results from snippet execution, ignored in this method
        :return: dict containing manually defined name / value pair
        """
        outputs = dict()

        if 'name' not in output_definition or 'capture_value' not in output_definition:
            return outputs

        var_name = output_definition['name']
        value = str(self.render(output_definition['capture_value'], self.context))

        outputs[var_name] = value

        return outputs

    def _handle_manual_outputs(self, results: str) -> dict:
        """
        Manually set a value in the context, this could be useful with 'when' conditionals

        :param results: results from snippet execution, ignored in this method
        :return: dict containing manually defined name / value pair
        """
        outputs = dict()

        try:
            if 'outputs' not in self.metadata:
                logger.info('No outputs defined in this snippet')
                return outputs

            for output in self.metadata['outputs']:

                if 'name' not in output:
                    continue

                var_name = output['name']
                value = output['value']

                outputs[var_name] = value

        except KeyError as ke:
            logger.error(f'Could not locate required attributes for manual output: {ke} in snippet: {self.name}')

        return outputs



import logging
from typing import List

from skilletlib.snippet.template import TemplateSnippet
from .base import Skillet

logger = logging.getLogger(__name__)


class TemplateSkillet(Skillet):
    snippet_required_metadata = {'name'}

    snippet_optional_metadata = {
        'file': '',
        'element': ''
    }

    def get_snippets(self) -> List[TemplateSnippet]:
        if hasattr(self, 'snippets'):
            return self.snippets

        snippet_list = list()
        for snippet_def in self.snippet_stack:
            if 'element' not in snippet_def or snippet_def['element'] == '':
                template_str = self.load_template(snippet_def.get('file', ''))
                snippet_def['element'] = template_str

            else:
                template_str = snippet_def['element']

            snippet = TemplateSnippet(template_str, snippet_def)
            snippet_list.append(snippet)

        return snippet_list

    def get_results(self) -> dict:
        """
        TemplateSkillet will add an additional attribute into the results dict containing the value of
        the first snippet found to have been successfully executed

        .. code-block:: json

            {
              "snippets": {
                "config_template": "success"
              },
              "template": "Rendered Template output"
            }


        :return: dict containing default outputs plus the rendered template contained in the 'template' attribute
        """

        results = super()._get_snippet_results()
        cleaned_results = dict()
        cleaned_results['snippets'] = dict()
        # include outputs for #104
        cleaned_results['outputs'] = self.captured_outputs
        snippets = results.get('snippets', {})
        for k, v in snippets.items():
            if v != '':
                cleaned_results['template'] = v['raw']
                cleaned_results['snippets'][k] = 'success'
                break
            cleaned_results['snippets'][k] = 'failure'

        return cleaned_results



from .panos import PanosSnippet


class PanValidationSnippet(PanosSnippet):
    """
    Pan validation Snippet
    """
    required_metadata = {'name'}
    # optional metadata that may be overridden in the snippet definition / metadata

    optional_metadata = {
        'fail_message': 'Snippet Validation Failed',
        'pass_message': 'Snippet Validation Passed'
    }

    def handle_output_type_validation(self, results: str):
        output = dict()
        output['results'] = results
        output['label'] = self.metadata.get('label', '')
        output['severity'] = self.metadata.get('severity', 'low')
        output['documentation_link'] = self.metadata.get('documentation_link', '')
        output['test'] = self.metadata.get('test', '')

        o = dict()
        o[self.name] = output
        return o



import logging
import shlex
import time
import traceback
from typing import Any
from typing import Tuple

from docker import DockerClient
from docker.errors import APIError
from docker.errors import ContainerError
from docker.errors import DockerException
from docker.errors import ImageNotFound

from skilletlib.exceptions import SkilletLoaderException
from .base import Snippet

logger = logging.getLogger(__name__)


class DockerSnippet(Snippet):
    """
    Docker Snippet - Represents a command to run inside a docker container
    """
    required_metadata = {'name', 'image', 'cmd'}

    # optional parameters that may be set in the snippet metadata
    optional_metadata = {
        'volumes': dict(),
        'async': True
    }

    template_metadata = {'cmd', 'tag'}

    output_type = 'text'

    # keep track of the last time we queried the logs
    last_logs_time = None

    def __init__(self, metadata):
        super().__init__(metadata)
        self.client = DockerClient()

        # configure from metadata
        image = self.metadata.get('image', 'python')
        if ':' in image:
            self.image, self.tag = image.split(':')

        else:
            self.image = image
            self.tag = 'latest'

        self.working_dir = self.metadata.get('working_dir', '/app')

        # this is set in the get_snippets method of the Docker Skillet class
        # and is used if no volume mounts have been specified
        self.path = self.metadata.get('skillet_path', None)

        self.detach = self.metadata.get('async', False)

        if self.detach:
            self.auto_remove = False
        else:
            self.auto_remove = True

        # set up volumes
        # A dictionary to configure volumes mounted inside the container.
        # The key is either the host path or a volume name, and the value is a dictionary with the keys:
        #     bind The path to mount the volume inside the container
        #     mode Either rw to mount the volume read/write, or ro to mount it

        volumes = self.metadata.get('volumes', dict())

        if not volumes:
            self.volumes = {self.path: {'bind': self.working_dir, 'mode': 'rw'}}

        else:
            self.volumes = volumes
            # if we have a volume passed in, ensure the working dir is set to the path of this skillet so
            # relative commands will work as intended
            self.working_dir = self.path

        # track our container
        self.container_id = ''

    def execute(self, context) -> Tuple[str, str]:
        """
        Execute this cmd in the specified docker container

        :param context: context containing all the user-supplied input variables. Also contains output from previous
        steps. Raises SkilletLoaderException on error
        :return:  Tuple(dict, str) output and string representing 'success' or 'failure'
        """
        try:

            output = ''
            logger.info(f'Pulling image: {self.image} with tag: {self.tag}')
            self.client.images.pull(self.image, self.tag)

            vols = self.volumes

            image = self.image + ":" + self.tag
            logger.info('Creating container...')
            return_data = self.client.containers.run(image, self.metadata['cmd'], volumes=vols, stderr=True,
                                                     detach=self.detach, working_dir=self.working_dir,
                                                     auto_remove=self.auto_remove, environment=context)

            if self.detach:
                # return_data will be a Container object if self.detach is True
                self.container_id = return_data.id
                output = self.container_id
                print('container id is ' + self.container_id)
                return output, 'running'

            else:
                # return_data will be the bytes returned from the command
                if type(return_data) is bytes:
                    return_str = return_data.decode('UTF-8')
                    return return_str, 'success'
                else:
                    return return_data, 'success'

        except ImageNotFound:
            logger.error(traceback.format_exc())
            raise SkilletLoaderException(f'Could not locate image {self.image} in {self.name}')
        except APIError as ae:
            logger.error(traceback.format_exc())
            raise SkilletLoaderException(f'Error communicating with Docker API: {ae}')
        except ContainerError as ce:
            logger.error(traceback.format_exc())
            raise SkilletLoaderException(f'Container command failed: {ce}')
        except DockerException as de:
            logger.error(traceback.format_exc())
            raise SkilletLoaderException(f'Could not execute docker container {self.name}: {de}')
        except ValueError as ve:
            # added or GL #77 - add diagnostics for failed docker container creation
            logger.error(traceback.format_exc())
            raise SkilletLoaderException(f'Could not execute docker container ValueError in {self.name}: {ve}')

    def get_container(self):
        try:
            return self.client.containers.get(self.container_id)
        except DockerException as de:
            logger.error(de)
            logger.error('Could not get container!')
            return None

    def get_output(self) -> Tuple[str, str]:
        if not self.detach:
            return '', 'success'

        try:
            container = self.get_container()

            if self.last_logs_time is None:
                return_data = container.logs()
            else:
                return_data = container.logs(since=self.last_logs_time)

            self.last_logs_time = int(time.time())

            return_str = self.__clean_output(return_data)

            if container.status == 'running':
                return return_str, 'running'

            else:
                logger.info(container.status)
                return return_str, 'success'

        except APIError as ae:
            raise SkilletLoaderException(f'Could not get logs for {self.name}: {ae}')

    def cleanup(self) -> None:
        """
        Clean up action is the docker container was started with 'async', no-op is async is False
        :return: None
        """
        if not self.detach:
            return

        try:
            container = self.get_container()

            if container:
                if container.status != 'running':
                    container.remove()
                else:
                    logger.warning(f'Docker container {self.container_id} may need to be manually removed!')

        except APIError as ae:
            raise SkilletLoaderException(f'Could not clean up {self.name}: {ae}')

    @staticmethod
    def __clean_output(return_data: Any) -> str:
        if type(return_data) is bytes:
            return_str = return_data.decode('UTF-8')
            return return_str
        else:
            return str(return_data)

    # Fix for GL #77 - escape cmd variables before passing into render_metadata
    def render_metadata(self, context: dict) -> dict:
        """
        render_metadata will ensure all metadata vars that contain jinja2 variables will be rendered properly.
        i.e. cmd: ansible-playbook -e 'somevar={{ some_value }}' will be converted to 'somevar='value_from_context'

        For docker snippets, we need to ensure each item in the context is properly escaped before being rendered

        :param context: context as provided from the user
        :return: fully rendered metadata
        """

        # only get the variables that are used in the 'cmd'
        affected_vars = self.get_variables_from_template(self.metadata['cmd'])

        # iterate over each var, get it's value from the context, quote it, and set back into the context
        for v in affected_vars:
            if v in context:
                context[v] = shlex.quote(context[v])

        return super().render_metadata(context)



# Copyright (c) 2018, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Authors: Nathan Embery

import datetime
import logging
import os
import random
import re
import sys
import time
from pathlib import Path
from typing import Optional
from typing import Tuple
from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError

import requests
import requests_toolbelt
import xmltodict
from lxml import etree
from lxml.etree import Element
from pan import xapi
from pan.config import PanConfig
from pan.xapi import PanXapiError
from xmldiff import main as xmldiff_main

from .exceptions import LoginException
from .exceptions import PanoplyException
from .exceptions import SkilletLoaderException
from .exceptions import TargetConnectionException
from .exceptions import TargetGenericException
from .exceptions import TargetLoginException
from .skilletLoader import SkilletLoader

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

if not len(logger.handlers):
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)


class Panoply:
    """
    Panoply is a wrapper around pan-python PanXAPI class to provide additional, commonly used functions
    """

    def __init__(self, hostname: Optional[str] = None, api_username: Optional[str] = None,
                 api_password: Optional[str] = None, api_port: Optional[int] = 443,
                 serial_number: Optional[str] = None, debug: Optional[bool] = False, api_key: Optional[str] = None):
        """
        Initialize a new panoply object. Passing in the authentication information will cause this class to attempt
        to connect to the device and set offline_mode to False. Otherwise, offline mode will be set to True

        :param hostname: hostname or ip address of target device`
        :param api_username: username
        :param api_password: password
        :param api_port: port to use for target device
        :param serial_number: Serial number of target device if proxy through panorama
        :param debug: Optional flag to log additional debug messages
        :param api_key: Optional api key to use instead of username / password auth
        """

        if api_port is None:
            api_port = 443

        self.hostname = hostname
        self.user = api_username
        self.pw = api_password
        self.port = api_port
        self.serial_number = serial_number
        self.key = api_key
        self.debug = debug

        self.serial = serial_number
        self.connected = False
        self.connected_message = 'no connection attempted'
        self.facts = {}
        self.last_error = ''
        self.offline_mode = False
        self.xapi = None

        if debug:
            logger.setLevel(logging.DEBUG)

        elif os.environ.get('SKILLET_DEBUG', False):
            logger.setLevel(logging.DEBUG)

        if hostname is None and api_username is None and api_password is None:
            self.offline_mode = True
            return

        # try to connect...
        self.offline_mode = False

        try:
            self.xapi = xapi.PanXapi(api_username=self.user, api_password=self.pw, hostname=self.hostname,
                                     port=self.port, serial=self.serial_number, api_key=self.key)

        except xapi.PanXapiError as pxe:
            err_msg = str(pxe)

            if '403' in err_msg:
                raise TargetLoginException('Invalid credentials logging into device')

            elif 'Errno 111' in err_msg:
                raise TargetConnectionException('Error contacting the device at the given IP / hostname')

            elif 'Errno 60' in err_msg:
                raise TargetConnectionException('Error contacting the device at the given IP / hostname')

            else:
                raise TargetGenericException(pxe)

        else:
            self.connect(allow_offline=True)

    def connect(self, allow_offline: Optional[bool] = False) -> None:
        """
        Attempt to connect to this device instance

        :param allow_offline: Do not raise an exception if this device is offline unless there is an authentication
        error
        :return: None
        """
        try:

            if self.xapi is None:
                self.xapi = xapi.PanXapi(api_username=self.user, api_password=self.pw, hostname=self.hostname,
                                         port=self.port, serial=self.serial_number)

            self.key = self.xapi.keygen()
            self.facts = self.get_facts()

        except PanXapiError as pxe:
            err_msg = str(pxe)

            if '403' in err_msg:
                raise LoginException('Invalid credentials logging into device')

            else:

                if allow_offline:
                    logger.debug('FYI - Device is not currently available')
                    self.connected = False
                    self.connected_message = 'device is not currently available'
                    return None

                else:
                    raise PanoplyException('Could not connect to device!')

        else:
            self.connected = True

    def commit(self, force_sync=True) -> str:
        """
        Perform a commit operation on this device instance -

        :raises PanoplyException: if commit failed
        :param force_sync: Flag to enable sync commit or async
        :return: String from the API indicating success or failure
        """

        try:
            self.xapi.commit(cmd='<commit></commit>', sync=force_sync, timeout=600)
            results = self.xapi.xml_result()

            if not self.__check_commit_return(results, force_sync):
                raise PanoplyException(self.xapi.status_detail)

            return self.xapi.status_detail

        except PanXapiError as pxe:
            logger.error(pxe)
            raise PanoplyException('Could not commit configuration')

    def commit_gpcs(self, force_sync=True) -> str:
        """
        Perform a commit operation on this device instance specifically for gpcs remote networks
        Note - you must do a full commit to panorama before you invoke this commit!

        :raises PanoplyException: if commit failed
        :param force_sync: Flag to enable sync commit or async
        :return: String from the API indicating success or failure
        """
        try:

            running_config = self.get_configuration(config_source='running')
            config_doc = etree.fromstring(running_config)

            sc = config_doc.xpath("/config/devices/entry[@name='localhost.localdomain']/"
                                  "device-group/entry[@name='Service_Conn_Device_Group']")
            rn = config_doc.xpath("/config/devices/entry[@name='localhost.localdomain']/"
                                  "device-group/entry[@name='Remote_Network_Device_Group']")
            mu = config_doc.xpath("/config/devices/entry[@name='localhost.localdomain']/"
                                  "device-group/entry[@name='Mobile_User_Device_Groupp']")

            if sc:
                self.xapi.commit(action='all',
                                 cmd='<commit-all><shared-policy><device-group>'
                                     '<entry name="Service_Conn_Device_Group"/>'
                                     '</device-group></shared-policy></commit-all>')

                results = self.xapi.xml_result()

                if not self.__check_commit_return(results, force_sync):
                    raise PanoplyException(self.xapi.status_detail)

            if rn:
                self.xapi.commit(action='all',
                                 cmd='<commit-all><shared-policy><device-group>'
                                     '<entry name="Remote_Network_Device_Group"/>'
                                     '</device-group></shared-policy></commit-all>')

                results = self.xapi.xml_result()

                if not self.__check_commit_return(results, force_sync):
                    raise PanoplyException(self.xapi.status_detail)

            if mu:
                self.xapi.commit(action='all',
                                 cmd='<commit-all><shared-policy><device-group>'
                                     '<entry name="Mobile_User_Device_Group"/>'
                                     '</device-group></shared-policy></commit-all>')

                results = self.xapi.xml_result()

                if not self.__check_commit_return(results, force_sync):
                    raise PanoplyException(self.xapi.status_detail)

            return 'Prisma Access Committed Successfully'

        except PanXapiError as pxe:
            logger.error(pxe)
            raise PanoplyException('Could not commit configuration')

    @staticmethod
    def __check_commit_return(results: str, force_sync: bool) -> bool:
        """
        Check xml result from a panos device and check for a failure condition

        :param results: str returned from xapi.xml_result
        :param force_sync: if force_sync is true, then verify the commit actually succeeded
        :return: boolean
        """

        if results is None:
            return False

        if force_sync:
            embedded_result = None
            try:
                doc = ElementTree.XML(results)
                embedded_result = doc.find('result')

            except ParseError:
                # results can be malformed xml from gpcs
                if 'jobid' in results:
                    return True

            if embedded_result is not None:
                commit_result = embedded_result.text

                if commit_result == 'FAIL':
                    return False

                else:
                    return True
        else:
            # fixme - is there something else we can check here?
            return True

    def set_at_path(self, name: str, xpath: str, xml_str: str) -> None:
        """
        Insert XML into the configuration tree at the specified xpath

        :param name: name of the snippet - used in logging and debugging only
        :param xpath: full xpath where the xml element will be inserted
        :param xml_str: string representation of the XML element to insert
        :return: None
        """

        try:
            self.xapi.set(xpath=xpath, element=self.sanitize_element(xml_str))

            if self.xapi.status_code == '7':
                raise SkilletLoaderException(f'xpath {xpath} was NOT found for skillet: {name}')

        except PanXapiError as pxe:
            raise PanoplyException(f'Could not push skillet {name} / snippet {xpath}! {pxe}')

    def execute_op(self, cmd_str: str, cmd_xml=False, parse_result=True) -> str:
        """
        Executes an 'op' command on the NGFW

        :param cmd_str: op command to send
        :param cmd_xml: Flag to determine if op command requires XML encoding
        :param parse_result: Optional flag to indicate whether to return parsed xml results (xml_result) from xapi - not
        all commands return valid XML. Setting this to 'false' will return the raw string from the API.
        :return: raw output from the device
        """

        try:
            self.xapi.op(cmd=cmd_str, cmd_xml=cmd_xml)
            if parse_result:
                return self.xapi.xml_result()
            else:
                return self.xapi.xml_document

        except PanXapiError as pxe:
            if 'ParseError' in str(pxe):
                # some op commands do not properly wrap data in CDATA tags, so let's try to
                # replace some chars manually and re-parse...
                x = self.xapi.xml_document
                cx = x.replace('&', '&amp;')

                try:
                    # make sure our little 'fix' didn't completely ruin the xml structure and return what we got
                    s = ElementTree.fromstring(cx)
                    return ElementTree.tostring(s)

                except Exception as e:
                    logger.error(e)
                    raise PanoplyException(pxe)

            raise PanoplyException(pxe)

    def execute_cli(self, cmd_str: str) -> str:
        """
        Short-cut to execute a simple CLI op cmd

        :param cmd_str: CLI command to send to the NGFW
        :return: raw output from the command
        """
        return self.execute_op(cmd_str, cmd_xml=True)

    def execute_cmd(self, cmd: str, params: dict, context=None) -> str:
        """
        Execute the given cmd using the xapi.

        :param cmd: Valid options are: 'op', 'show', 'get', 'delete', 'set', 'edit', 'override', 'move', 'rename',
                                       'clone', 'validate'
        :param params: valid parameters for the given cmd type
        :param context: skillet context
        :return: raw results from the cmd output, raises SkilletLoaderException
        """
        if cmd not in ('op', 'set', 'edit', 'override', 'move', 'rename', 'clone', 'show', 'get', 'delete'):
            raise PanoplyException('Invalid cmd type given to execute_cmd')

        # this code happily borrowed from ansible-pan module
        # https://raw.githubusercontent.com/PaloAltoNetworks/ansible-pan/develop/library/panos_type_cmd.py

        func = getattr(self.xapi, cmd)

        # shortcut for op cmd types
        if cmd == 'op':
            cmd_str = ''.join(params['cmd_str'].strip().split('\n'))

            # fix for GL #79
            parse_result = params.get('parse_result', True)

            return self.execute_op(cmd_str, cmd_xml=False, parse_result=parse_result)

        # in all other cases, the xpath is a required attribute
        kwargs = {
            'xpath': ''.join(params['xpath'].strip().split('\n')),
            'extra_qs': params.get('extra_qs', dict())
        }

        try:

            if cmd in ('set', 'edit', 'override'):
                kwargs['element'] = params['element'].strip()

            if cmd in ('move',):
                kwargs['where'] = params['where']
                # dst is optional
                kwargs['dst'] = params.get('dst', None)

            if cmd in ('rename', 'clone'):

                if 'new_name' in params:
                    kwargs['newname'] = params['new_name']

                else:
                    kwargs['newname'] = params['newname']

            if cmd in ('clone',):
                kwargs['xpath_from'] = params['xpath_from']

        except KeyError as ke:
            raise PanoplyException(f'Invalid parameters passed to execute_cmd: {ke}')

        try:
            func(**kwargs)

        except PanXapiError as e:
            raise PanoplyException(f'Could not execute command: {cmd}: {e}')

        return self.xapi.xml_result()

    def fetch_license(self, auth_code: str) -> bool:
        """
        Fetch and install licenses for PAN-OS NGFW

        :param auth_code: Authorization code to use to license the NGFW
        :return: True when license installation succeeds / False otherwise
        """

        if self.facts.get('family', '') == 'vm':

            if self.facts.get('vm-license', 'none') != 'none':
                logger.debug('This VM is already licensed')
                return True

        try:
            cmd = f'<request><license><fetch><auth-code>{auth_code}</auth-code></fetch></license></request>'
            logger.debug(f'Using request cmd: {cmd}')
            self.xapi.op(cmd=cmd)
            results = self.xapi.xml_result()
            logger.debug(f'fetch_license results: {results}')

            if 'License installed' in results:
                return True

            else:
                logger.warning('Unexpected Results from fetch_license')
                logger.warning(results)
                return False

        except PanXapiError as pxe:
            # bug present in 9.0.4 that returns content-type of xml but the content is only text.
            # this causes a ParseError to be thrown, however, the operation was actually successful

            if 'ParseError' in str(pxe):

                if self.xapi.xml_document is not None and 'License installed' in self.xapi.xml_document:
                    return True

            logger.error(f'Caught Exception in fetch_license: {pxe}')

            if self.xapi.status_detail:
                logger.error(self.xapi.status_detail)

            return False

    def set_license_api_key(self, api_key: str) -> bool:
        """
        Set's the Palo Alto Networks Support API Key in the firewall. This is required to deactivate a VM-Series NGFW.

        :param api_key: Your support account API Key found on the support.paloaltonetworks.com site
        :return: boolean True on success / False otherwise
        """

        try:

            try:
                verify_cmd = '<request><license><api-key><show></show></api-key></license></request>'
                verify_results = self.execute_op(verify_cmd)
                logger.debug(verify_results)
                current_api_key = verify_results.split(': ')[1]

                if current_api_key == api_key:
                    logger.debug('API Key is already set')
                    return True

            except PanoplyException as pxe:
                # will raise an error if there is no key already set, just log it and continue
                logger.debug(pxe)

            cmd = f'<request><license><api-key><set><key>{api_key}</key></set></api-key></license></request>'
            logger.debug(f'Using request cmd: {cmd}')
            results = self.execute_op(cmd)
            logger.debug(f'set_license_api_key results: {results}')

            if 'successfully set' in results:
                return True

            else:
                logger.warning('Unexpected Results from set_license_api_key')
                logger.warning(results)
                return False

        except PanoplyException as pxe:
            logger.error(f'Caught Exception in set_license_api_key: {pxe}')

            if self.xapi.status_detail:
                logger.error(self.xapi.status_detail)

            return False

    def deactivate_vm_license(self, api_key: str = None) -> bool:
        """
        Deactivate VM-Series Licenses. Will set the API Key is not already set.

        :param api_key: Optional api_key.
        :return: boolean True on success / False otherwise
        """

        if self.facts.get('family', '') != 'vm':
            logger.warning('Attempting to deactivate license on a NGFW that is not a VM!')
            return False

        if self.facts.get('vm-license', 'none') == 'none':
            logger.debug('This VM is already de-licensed')
            return True

        if api_key is not None:

            if not self.set_license_api_key(api_key):
                return False

        try:
            cmd = '<request><license><deactivate><VM-Capacity><mode>auto</mode>' \
                  '</VM-Capacity></deactivate></license></request>'
            logger.debug(f'Using request cmd: {cmd}')
            results = self.execute_op(cmd)
            logger.debug(f'deactivate_vm_license results: {results}')

            if 'Successfully deactivated' in results:
                return True

            else:
                logger.warning('Unexpected Results from deactivate_vm_license')
                logger.warning(results)
                return False

        except PanoplyException as pxe:
            logger.error(f'Caught Exception in deactivate_vm_license: {pxe}')

            if self.xapi.status_detail:
                logger.error(self.xapi.status_detail)

            return False

    @staticmethod
    def sanitize_element(element: str) -> str:
        """
        Eliminate some unneeded characters from the XML snippet if they appear.

        :param element: element str
        :return: sanitized element str
        """
        element = re.sub(r"\n\s+", "", element)
        element = re.sub(r"\n", "", element)

        return element

    def backup_config(self):
        """
        Saves a named backup on the PAN-OS device. The format for the backup is 'panhandler-20190424000000.xml'

        :return:  xml results from the op command sequence
        """
        d = datetime.datetime.today()
        tstamp = d.strftime('%Y%m%d%H%M%S')
        cmd = f'<save><config><to>config-backup-{tstamp}.xml</to></config></save>'
        try:
            self.xapi.op(cmd=cmd)

            if self.xapi.status != 'success':
                raise PanoplyException('Could not perform config backup on device!')

            return self.xapi.xml_result()

        except xapi.PanXapiError as pxe:
            raise PanoplyException(f'Could not perform backup: {pxe}')

    def get_facts(self) -> dict:
        """
        Gather system info and keep on self.facts
        This gets called on every connect

        :return: dict containing all system facts
        """
        facts = {}

        # FIXME - add better error handling here
        self.xapi.op(cmd='<show><system><info></info></system></show>')

        if self.xapi.status != 'success':
            raise PanoplyException('Could not get facts from device!')

        results_xml_str = self.xapi.xml_result()
        results = xmltodict.parse(results_xml_str)

        if 'system' in results:
            facts.update(results['system'])

        self.xapi.show(xpath="./devices/entry[@name='localhost.localdomain']/deviceconfig/system")
        results_xml_str = self.xapi.xml_result()
        results = xmltodict.parse(results_xml_str)

        if 'system' in results:
            facts['timezone'] = results['system'].get('timezone', 'US/Pacific')

        try:
            facts['dns-primary'] = results['system']['dns-setting']['servers']['primary']
            facts['dns-secondary'] = results['system']['dns-setting']['servers']['secondary']

        except KeyError:
            # DNS is not configured on the host, but we will need it later for some noob operations
            facts['dns-primary'] = '1.1.1.1'
            facts['dns-secondary'] = '1.0.0.1'

        return facts

    def load_baseline(self) -> bool:
        """
        Load baseline config that contains ONLY connecting username / password
        use device facts to determine which baseline template to load
        see template/panos/baseline_80.xml for example

        :param self:
        :return: bool true on success
        """

        file_contents = self.generate_baseline()
        self.import_file('skillet_baseline.xml', file_contents, 'configuration')

        return self.load_config('skillet_baseline.xml')

    def generate_baseline(self, reset_hostname=False) -> str:
        """
        Load baseline config that contains ONLY connecting username / password
        use device facts to determine which baseline template to load
        see template/panos/baseline_80.xml for example

        :param self:
        :param reset_hostname: Flag to reset hostname back to a default value (baseline in this case)
        :return: string contents of baseline config
        """

        if not self.connected:
            self.connect()

        if 'sw-version' not in self.facts:
            raise PanoplyException('Could not determine sw-version to load baseline configuration!')

        version = self.facts['sw-version']
        context = dict()
        context['ADMINISTRATOR_USERNAME'] = self.user
        context['ADMINISTRATOR_PASSWORD'] = self.pw
        # fix for #50 - always set DNS values
        context['DNS_1'] = self.facts['dns-primary']
        context['DNS_2'] = self.facts['dns-secondary']

        if self.facts['model'] == 'Panorama':
            skillet_type_dir = 'panorama'

            if not reset_hostname:
                context['PANORAMA_NAME'] = self.facts['hostname']
            else:
                context['PANORAMA_NAME'] = 'baseline'

            # FIXME - is there a way to determine if dhcp is active via an op cmd?
            context['PANORAMA_TYPE'] = 'static'
            context['PANORAMA_IP'] = self.facts['ip-address']
            context['PANORAMA_MASK'] = self.facts['netmask']
            context['PANORAMA_DG'] = self.facts['default-gateway']

        else:
            skillet_type_dir = 'panos'

            if not reset_hostname:
                context['FW_NAME'] = self.facts['hostname']
            else:
                context['FW_NAME'] = 'baseline'

            if self.facts['is-dhcp'] == 'no':
                context['MGMT_TYPE'] = 'static'
                context['MGMT_IP'] = self.facts['ip-address']
                context['MGMT_MASK'] = self.facts['netmask']
                context['MGMT_DG'] = self.facts['default-gateway']

        if '8.0' in version:
            # load the 8.0 baseline with
            skillet_dir = 'baseline_80'

        elif '8.1' in version:
            # load the 8.1 baseline with
            skillet_dir = 'baseline_81'

        elif '9.0' in version:
            # load the 9.0 baseline with
            skillet_dir = 'baseline_90'

        elif '9.1' in version:
            # load the 9.1 baseline with
            skillet_dir = 'baseline_91'

        # Support 9.2 Beta and 10.0 for GL #80
        elif '9.2' in version:
            # load the 9.1 baseline with
            skillet_dir = 'baseline_91'

        elif '10.0' in version:
            # load the 9.1 baseline with
            skillet_dir = 'baseline_91'

        else:
            raise PanoplyException('Could not determine sw-version for baseline load')

        template_path = Path(__file__).parent.joinpath('assets', skillet_type_dir, skillet_dir)
        sl = SkilletLoader()
        baseline_skillet = sl.load_skillet_from_path(str(template_path.resolve()))
        snippets = baseline_skillet.get_snippets()
        snippet = snippets[0]
        output, status = snippet.execute(context)

        if status == 'success':
            return str(output)

        else:
            raise PanoplyException('Could not generate baseline config!')

    def import_file(self, filename: str, file_contents: (str, bytes), category: str) -> bool:
        """
        Import the given file into this device

        :param filename:
        :param file_contents:
        :param category: 'configuration'
        :return: bool True on success
        """
        params = {
            'type': 'import',
            'category': category,
            'key': self.key
        }

        mef = requests_toolbelt.MultipartEncoder(
            fields={
                'file': (filename, file_contents, 'application/octet-stream')
            }
        )

        r = requests.post(
            f'https://{self.hostname}:{self.port}/api/',
            verify=False,
            params=params,
            headers={'Content-Type': mef.content_type},
            data=mef
        )

        # if something goes wrong just raise an exception
        r.raise_for_status()

        resp = ElementTree.fromstring(r.content)

        if resp.attrib['status'] == 'error':
            raise PanoplyException(r.content)

        return True

    def load_config(self, filename: str) -> bool:
        """
        Loads the named configuration file into this device

        :param filename: name of the configuration file on the device to load. Note this filename must already exist
        on the target device
        :return: bool True on success
        """

        cmd = f'<load><config><from>{filename}</from></config></load>'
        self.xapi.op(cmd=cmd)

        if self.xapi.status == 'success':
            return True

        else:
            return False

    def has_running_jobs(self) -> bool:
        """
        Simple check to determine if there are any running jobs on this device

        :return: bool True if there is currently a running job
        """

        self.execute_cli('show jobs all')
        jobs = self.xapi.xml_document
        jobs_element = etree.fromstring(jobs)

        running_jobs_list = jobs_element.xpath(".//jobs/status[text() != 'FIN']")

        return True if running_jobs_list else False

    def wait_for_device_ready(self, interval=30, timeout=600) -> bool:
        """
        Loop and wait until device is ready or times out

        .. note::

            This currently only supports PAN-OS devices and not Panorama

        :param interval: how often to check in seconds
        :param timeout: how long to wait until we declare a timeout condition
        :return: boolean true on ready, false on timeout
        """
        mark = time.time()
        timeout_mark = mark + timeout

        while True:
            try:

                # fix for #60 - show chassis ready is not available on panorama
                if self.facts.get('model', 'Panorama') == 'Panorama':
                    cmd = "<show><system><info></info></system></show>"
                    is_panorama = True
                else:
                    cmd = '<show><chassis-ready></chassis-ready></show>'
                    is_panorama = False

                self.xapi.op(cmd=cmd)
                resp = self.xapi.xml_result()

                if self.xapi.status == 'success':
                    # in the case of panorama, we may be up but the auto-commit job may still be running
                    # continue to wait until there are no more jobs
                    # FIXME - should probably enhance this to only check for auto-commit job, it's possible there
                    # can be other jobs running with a busy / heavily used panorama instance
                    if is_panorama:
                        if not self.has_running_jobs():
                            return True

                    if resp.strip() == 'yes':
                        return True

            except TargetLoginException:
                logger.error('Could not log in to device...')
                return False

            except PanXapiError:
                logger.info(f'{self.hostname} is not yet ready...')

            if time.time() > timeout_mark:
                return False

            logger.info(f'Waiting for {self.hostname} to become ready...')
            time.sleep(interval)

    def filter_connected_devices(self, filter_terms=None) -> list:
        """
        Returns the list of connected devices filtered according to the given terms.

        Filter terms are based on keys in the device facts. Matches are done using simple regex match

        :param filter_terms: dict containing key value pairs. Keys match keys from the return device facts and values
        are regex expressions used to match those keys
        :return: list of devices that match ALL filter terms
        """

        if filter_terms is None:
            filter_terms = {}

        if self.facts.get('model', '') != 'Panorama':
            logger.info('Method only supported on Panorama Devices')
            return list()

        connected_devices_xml = self.execute_cli('show devices connected')
        connected_devices_dict = xmltodict.parse(connected_devices_xml)

        if 'devices' in connected_devices_xml and 'entry' in connected_devices_dict['devices']:
            connected_devices = connected_devices_dict['devices']['entry']

        else:
            return list()

        if not filter_terms:
            # no terms given, return all devices
            return connected_devices

        filtered_devices = list()

        for device in connected_devices:
            match = False

            for term in filter_terms:
                if term in device:

                    if re.match(filter_terms[term], device[term]):
                        match = True
                    else:
                        match = False

            if match:
                filtered_devices.append(device)

        return filtered_devices

    def update_dynamic_content(self, content_type: str) -> bool:
        """
        Check for newer dynamic content and install if found

        :param content_type: type of content to check. can be either: 'content', 'anti-virus', 'wildfire'
        :return: bool True on success
        """
        try:
            version_to_install = self.check_content_updates(content_type)

            if version_to_install is None:
                logger.info(f'Latest {content_type} version is already installed')
                return True

            logger.info(f'Downloading Dynamic Content of type: {content_type}')
            cmd = f'<request>' \
                  f'<{content_type}><upgrade><download><latest/></download></upgrade></{content_type}>' \
                  f'</request>'

            self.xapi.op(cmd=cmd)
            results_element = self.xapi.element_result
            job_element = results_element.find('.//job')

            if job_element is not None:
                job_id = job_element.text

                if not self.wait_for_job(job_id):
                    raise PanoplyException('Could not update dynamic content')

            logger.info(f'Installing newer {content_type}')
            install_cmd = f'<request><{content_type}><upgrade><install>' \
                          f'<version>latest</version><commit>no</commit></install></upgrade></{content_type}></request>'

            self.xapi.op(cmd=install_cmd)
            results_element = self.xapi.element_result
            job_element = results_element.find('.//job')

            if job_element is not None:
                job_id = job_element.text

                if not self.wait_for_job(job_id):
                    raise PanoplyException('Could not install dynamic content')

            else:
                logger.info('No job returned to track')

            return True

        except PanXapiError:
            logger.error('Could not check for updated dynamic content')
            return False

    def check_content_updates(self, content_type: str) -> (str, None):
        """
        Iterate through all available content of the specified type, locate and return the version with the highest
        version number. If that version is already installed, return None as no further action is necessary

        :param content_type: type of content to check
        :return: version-number to download and install or None if already at the latest
        """
        latest_version = ''
        latest_version_first = 0
        latest_version_second = 0
        latest_version_current = 'no'

        try:
            logger.info(f'Checking for newer {content_type}...')
            self.xapi.op(cmd=f'<request><{content_type}><upgrade><check/></upgrade></{content_type}></request>')
            er = self.xapi.element_root

            for entry in er.findall('.//entry'):
                version = entry.find('./version').text
                current = entry.find('./current').text
                # version will have the format 1234-1234
                version_parts = version.split('-')
                version_first = int(version_parts[0])
                version_second = int(version_parts[1])

                if version_first > latest_version_first and version_second > latest_version_second:
                    latest_version = version
                    latest_version_first = version_first
                    latest_version_second = version_second
                    latest_version_current = current

            if latest_version_current == 'yes':
                return None

            else:
                return latest_version

        except PanXapiError:
            return None

    def wait_for_job(self, job_id: str, interval=10, timeout=600) -> bool:
        """
        Loops until a given job id is completed. Will timeout after the timeout period if the device is
        offline or otherwise unavailable.

        :param job_id: id the job to check and wait for
        :param interval: how long to wait between checks
        :param timeout: how long to wait with no response before we give up
        :return: bool true on content updated, false otherwise
        """
        mark = time.time()
        timeout_mark = mark + timeout
        logger.debug(f'Waiting for job id: {job_id} to finish...')
        while True:

            try:
                self.xapi.op(cmd=f'<show><jobs><id>{job_id}</id></jobs></show>')

            except PanXapiError as pxe:
                logger.error(f'Exception checking job: {pxe}')
                logger.error(f'Could not locate job with id: {job_id}')
                return False

            if self.xapi.status == 'success':
                job_element = self.xapi.element_result
                job_status_element = job_element.find('.//status')

                if job_status_element is not None:
                    job_status = job_status_element.text

                    if job_status == 'FIN':
                        logger.debug(f'Job: {job_id} is now complete')
                        return True

                    elif job_status == 'ACT':
                        job_progress_element = job_element.find('.//progress')

                        if job_progress_element is not None:
                            job_progress = job_progress_element.text
                            logger.debug(f'Progress is currently: {job_progress}')

                else:
                    logger.error('No job status element to be found!')
                    return False

            else:
                logger.debug(f'{self.xapi.xml_result()}')

                if time.time() > timeout_mark:
                    return False

                logger.info('Waiting a bit longer')

            time.sleep(interval)

    def get_configuration(self, config_source='running') -> str:
        """
        Get the configuration from the device.

        :return: configuration xml as a string or a blank string if not connected
        """

        if config_source == 'baseline':
            return self.generate_baseline()

        elif config_source == 'candidate':
            cmd = 'show config candidate'

        else:
            cmd = 'show config running'

        try:

            if self.connected:
                self.xapi.op(cmd=cmd, cmd_xml=True)
                return self.xapi.xml_result()

            else:
                return ''

        except PanXapiError:
            logger.error('Could not get configuration from device')
            raise PanoplyException('Could not get configuration from the device')

    def get_saved_configuration(self, configuration_name: str) -> str:
        """
        Returns a saved configuration on the device. Use 'list_saved_configuration' to get a list of available options

        :param configuration_name: name of the saved configuration to export
        :return: configuration as an XML encoded string
        """

        try:

            if self.connected:
                self.xapi.op(f'<show><config><saved>{configuration_name}</saved></config></show>', cmd_xml=False)

                if self.xapi.status != 'success':
                    raise PanoplyException('Could not get saved configuration from the device')

                doc_str = self.xapi.xml_result()

                doc = etree.fromstring(doc_str)
                return etree.tostring(doc, pretty_print=True).decode(encoding='UTF-8')

            else:
                return ''

        except PanXapiError:
            logger.error('Could not get saved configuration from device')
            raise PanoplyException('Could not get saved configuration from the device')

    def list_saved_configurations(self) -> list:
        """
        Returns a list of saved configuration files on this device

        :return: list of saved configurations
        """

        saved_configurations = list()
        try:
            if not self.connected:
                return saved_configurations

            self.xapi.ad_hoc(qs='type=op&action=complete&xpath=/operations/load/config/from', modify_qs=True)

            if self.xapi.status != 'success':
                raise PanoplyException('Could not list saved configurations from the device')

            response_root = self.xapi.element_root
            config_items = response_root.findall('.//completion')
            for config in config_items:
                if 'value' in config.attrib:
                    saved_configurations.append(config.attrib['value'])

            return saved_configurations

        except PanXapiError as pe:
            logger.error('Could not list configurations from device')
            logger.error(pe)
            raise PanoplyException('Could not list saved configuration from the device')

    def generate_skillet(self, from_candidate=False) -> list:
        """
        Generates a skillet from the changes detected on this device.
        This will attempt to create the xml and xpaths for everything that is found to have changed

        :param from_candidate: If your changes on in the candidate config, this will detect changes between the running
        config and the candidate config. If False, this will detect changes between the running config and a generic
        baseline configuration
        :return: list of xpaths
        """

        if from_candidate:
            self.xapi.op(cmd='show config candidate', cmd_xml=True)
            latest_config = self.xapi.xml_result()
            self.xapi.op(cmd='show config running', cmd_xml=True)
            previous_config = self.xapi.xml_result()

        else:
            previous_config = self.generate_baseline(reset_hostname=True)
            self.xapi.op(cmd='show config running', cmd_xml=True)
            latest_config = self.xapi.xml_result()

        return self.generate_skillet_from_configs(previous_config, latest_config)

    def generate_skillet_from_configs(self, previous_config: str, latest_config: str) -> list:
        """
        Generates a skillet from the diffs between two XML configuration files

        :param previous_config: Configuration backup taken before desired changes are made to the device
        :param latest_config: Configuration backup taken after desired changes are made

        :return: list of dictionaries that contain the following keys:
            * name
            * element
            * xpath
            * full_xpath
        """
        # convert the config string to an xml doc
        latest_doc = etree.fromstring(latest_config)

        # let's grab the previous as well
        previous_doc = etree.fromstring(previous_config)

        current_xpath = '.'
        not_found_xpaths = list()

        for c in latest_doc:
            o_xpath = current_xpath + '/' + c.tag
            these_not_found_xpaths = self.__check_element(c, o_xpath, previous_doc, [])
            not_found_xpaths.extend(these_not_found_xpaths)

        snippets = list()

        for xpath in not_found_xpaths:

            # xpath comes as a relative full xpath like './mgt-config/password-complexity'
            # make it /config/mgt-config/password-complexity
            full_xpath = re.sub(r'^\./', r'/config/', xpath)
            # force full xpath for #52 - split off last bit like password-complexity
            set_xpath, entry = self.__split_xpath(full_xpath)
            # remove any attribute for tag name
            tag = re.sub(r'\[.*\]', '', entry)

            # check if this xpath is actually user configurable
            if self.__is_ignored_xpath(set_xpath):
                continue

            # in some cases we need to ignore xpath + element for example
            # /config/shared/content-preview
            if self.__is_ignored_xpath(full_xpath):
                continue

            changed_elements = latest_doc.xpath(xpath)

            if not changed_elements:
                print('Something is broken here')
                continue

            changed_element = changed_elements[0]

            cleaned_element = self.__clean_uuid(changed_element)
            xml_string = etree.tostring(cleaned_element, pretty_print=True).decode(encoding='UTF-8')

            random_name = str(int(random.random() * 1000000))

            snippet = dict()
            snippet['name'] = f'{tag}-{random_name}'
            snippet['xpath'] = set_xpath
            snippet['element'] = xml_string.strip()
            snippet['full_xpath'] = xpath
            snippets.append(snippet)

        return self.__order_snippets(snippets)

    @staticmethod
    def __is_ignored_xpath(set_xpath: str) -> bool:
        """
        Determines if the given set_xpath should be used for skillet generation purposes. Certain xpaths will be
        changed, however, they are not user configurable, and therefore should not be included

        :param set_xpath: xpath string to check
        :return: boolean true if the xpath should be skipped
        """
        ignored_xpaths = (
            '/config/mgt-config/users/entry[@name="admin"]',
            '/config/shared/content-preview',
            '/config/readonly'
        )

        for ignored_xpath in ignored_xpaths:
            if ignored_xpath in set_xpath:
                logger.debug(f'Skipping ignored xpath: {set_xpath}')
                return True

        return False

    @staticmethod
    def __is_ignored_set_cli(set_cli: str) -> bool:
        """
        Determines if the generated set cli can be used to recreate this configuration. Some generated
        set clis are simply remnants of the XML structure and are not actually valid set cli commands

        :param set_cli: set command to check
        :return: bool True if this cli should be excluded
        """

        # ignore any set clis that start with this commands
        ignored_set_prefix = (
            'set mgt-config users admin phash',
            'set shared content-preview',
            'set read-only'
        )

        # ignore any set cli that match these commands
        ignored_set_cli = (
            'set shared application',
            'set shared application-group',
            'set shared service',
            'set shared service-group'
        )

        # ignore any set cli that contain these strings
        ignored_set_cli_parts = (
            'deviceconfig setting management initcfg',
            'deviceconfig setting management disable-predefined-reports'

        )

        for prefix in ignored_set_prefix:
            if set_cli.startswith(prefix):
                logger.debug(f'Skipping ignore set cli prefix: {set_cli}')
                return True

        for ignorned_set in ignored_set_cli:
            if set_cli == ignorned_set:
                logger.debug(f'Skipping ignore set cli: {set_cli}')
                return True

        for part in ignored_set_cli_parts:
            if part in set_cli:
                logger.debug(f'Skipping ignored set cli with part: {set_cli}')
                return True

        return False

    def generate_set_cli_from_configs(self, previous_config: str, latest_config: str) -> list:
        """
        Takes two configuration files, converts them to set commands, then returns only the commands found
        in the 'latest_config' vs the 'previous_config'. This allows the user to quickly configure one firewall,
        generate a 'set cli' diff and move those configs to another firewall

        :param previous_config: Starting config
        :param latest_config: Ending config
        :return: list of set cli commands required to convert previous to latest
        """

        p_config = PanConfig(previous_config)
        l_config = PanConfig(latest_config)

        p_set = p_config.set_cli('set ', xpath='./')
        l_set = l_config.set_cli('set ', xpath='./')

        # diffs = [item for item in l_set if item not in p_set]
        diffs = list()

        for cmd in l_set:

            if cmd not in p_set:

                if self.__is_ignored_set_cli(cmd):
                    continue

                cmd_cleaned = cmd.replace('\n', ' ')
                diffs.append(cmd_cleaned)

        return self.__order_set_commands(diffs)

    def __check_element(self, el: etree.Element, xpath: str, pc: etree.Element, not_founds: list) -> list:
        """
        recursive function to determine if the 'el' Element found at 'xpath' can also be found at the same
        xpath in the 'pc' previous_config. Keep tabs on what has not been found using the 'not_founds' list

        :param el: The element in question from the latest_config
        :param xpath: the xpath to the element in question
        :param pc:  the previous config Element
        :param not_founds: list of xpaths that have not been found
        :return: a list of xpaths that have not been found in the previous_config at this level
        """

        # first, check the previous_config to see if this xpath exists there
        found_elements = pc.xpath(xpath)

        if found_elements:
            found_element = found_elements[0]
            # this xpath exists in the previous_config, now iterate through all the children
            children = el.findall('./')

            if len(children) == 0:
                # there are no children at this level, so check if the 'text' is different

                if found_element.text != el.text:
                    # this xpath contains a text node that has been modified <port>6666</port> != <port>0000</port>
                    not_founds.append(xpath)
                # no need to go further as we have no children to descend into
                return not_founds

            # we have children elements, first check if they are a list of identical elements
            is_list = False

            if self.__check_children_are_list(children):
                # use the xmldiff library to check the list of elements
                diffs = xmldiff_main.diff_trees(found_element, el,
                                                {'F': 0.1, 'ratio_mode': 'accurate', 'fast_match': True})

                # all children are a list and there are no differences in them, so return up the stack
                if len(diffs) == 0:
                    return not_founds

                else:
                    # we have a list and there ARE differences
                    is_list = True

            # continue checking each child, either they are not a list or they are a list and there are diffs
            # track the child index in case we find a diff in the list case
            index = 1
            # check each child now
            for e in el:

                if e.attrib:
                    attribs = list()

                    for k, v in e.attrib.items():

                        if k != 'uuid':
                            attribs.append(f'@{k}="{v}"')

                    # fix for #71
                    attrib_str = " and ".join(attribs)
                    # track the attributes in the xpath by virtue of the 'path_entry' which will be appended to the
                    # xpath later
                    path_entry = f'{e.tag}[{attrib_str}]'

                else:
                    # no attributes but this is a list, so include the index value in the xpath to check
                    # this will be used to grab the changed element later, but will be removed from the xpath
                    # as it is not necessary in PAN-OS (double check this please)

                    if is_list:

                        if e.text.strip() != '':
                            path_entry = f'{e.tag}[text()="{e.text.strip()}"]'

                        else:
                            path_entry = f'{e.tag}[{index}]'

                    else:
                        # just append the tag to the xpath and move on
                        path_entry = e.tag

                # craft our new xpath to check
                n_xpath = xpath + '/' + path_entry
                # do it all over again
                new_not_founds = self.__check_element(e, n_xpath, pc, list())
                # add any child xpaths that weren't found with any found here for return up the stack
                not_founds.extend(new_not_founds)
                # increase our index for the next iteration
                index = index + 1

            # return our findings up the stack
            return not_founds

        # check this element to determine if it's 'blank'
        if len(el.findall('./')) == 0 and not el.attrib and (not el.text or not el.text.strip()):
            return not_founds

        not_founds.append(xpath)
        return not_founds

    @staticmethod
    def __split_xpath(xpath: str) -> Tuple[str, str]:
        xpath_parts = re.split(r'/(?=(?:[^"]*"[^"]*")*[^"]*$)', xpath)
        new_xpath = "/".join(xpath_parts[:-1])
        entry = xpath_parts[-1]

        return new_xpath, entry

    def __order_snippets(self, snippets: list) -> list:
        """
        Orders a set of snippets according to PAN-OS rules. This is an evolving list, which will almost certainly
        be 100% correct. Please file issues if you find an ordering issue in auto-generated snippets or set commands

        :param snippets: list of snippet dictionaries to order
        :return:  list of a snippet dictionary in the order in which they need to be applied to a NGFW or Panorama
        """
        # Attempt to order the snippets in a cohesive ordering. Will never be 100% perfect,
        # but at least make the attempt

        xpaths, post_xpaths = self.get_ordered_xpaths()

        ordered_snippets = list()
        for x in xpaths:
            found_snippets = self.__filter_snippets_by_xpath(snippets, x)
            for s in found_snippets:
                if s not in ordered_snippets:
                    ordered_snippets.append(s)

        post_snippets = list()
        for p in post_xpaths:
            filtered_snippets = self.__filter_snippets_by_xpath(snippets, p)
            for f in filtered_snippets:
                if f not in post_snippets:
                    post_snippets.append(f)

        for s in snippets:
            if s not in ordered_snippets and s not in post_snippets:
                ordered_snippets.append(s)

        for ps in post_snippets:
            if ps not in ordered_snippets:
                ordered_snippets.append(ps)

        return ordered_snippets

    def __order_set_commands(self, set_commands: list) -> list:
        """
        This will attempt to order a list of set commands using the same logic as ordering snippets.
        This is done by converting the set command into a 'fake' xpath, creating a fake snippet dictionary
        and sending those through the order_snippets method

        :param set_commands: list of set commands to order
        :return: a list of ordered set commands
        """

        ordered_set_commands = list()

        fake_snippets = list()

        for set_cmd in set_commands:
            snippet = dict()
            snippet['full_xpath'] = set_cmd.replace(' ', '/')
            fake_snippets.append(snippet)

        ordered_fake_snippets = self.__order_snippets(fake_snippets)

        for fake_snippet in ordered_fake_snippets:
            new_set_command = fake_snippet['full_xpath'].replace('/', ' ')
            ordered_set_commands.append(new_set_command)

        return ordered_set_commands

    @staticmethod
    def __filter_snippets_by_xpath(snippets: list, xpath: str) -> list:
        """
        Check the each snippet in the list and return those that has a full_xpath matching the xpath param. This is
        done by finding the specific leaf node of the xpath by using only the xpath portions after any template, device,
        or vsys entries. So an xpath pattern like 'network/interface' will match even within templates, across vsys,
        etc

        :param snippets: list of snippets
        :param xpath: xpath to check
        :return: list of only those that contain the xpath in their full_xpath attribute
        """
        filtered_snippets = list()

        split_pattern = re.compile(r'/devices/.*?/|vsys/.*?/')

        for s in snippets:
            full_xpath = s.get('full_xpath', '')

            split_xpath = split_pattern.split(full_xpath)
            leaf_xpath_initial = split_xpath[-1]

            # handle cases like ./mgt-config/password-complexity - remove the leading './'
            leaf_xpath = re.sub(r'^\./', '', leaf_xpath_initial)
            if leaf_xpath.startswith(xpath):
                # note we do not remove found snippets from the source snippets list, which may result
                # in duplicates. The calling code will need to ensure it does not append the results of this method
                # which out checking for dups first
                filtered_snippets.append(s)

        return filtered_snippets

    @staticmethod
    def __check_children_are_list(c: list) -> bool:
        # check if children are a list of items by identical tag names

        if len(c) <= 1:
            # can't be a list of identical items if there are only 0 or 1 items
            return False

        found_tag_name = ''

        for child in c:

            if found_tag_name == '':
                found_tag_name = child.tag
                continue

            if found_tag_name != child.tag:
                return False

        return True

    def generate_skillet_from_configs_old(self, previous_config: str, latest_config: str) -> list:
        # use the excellent xmldiff library to get a list of changed elements
        diffs = xmldiff_main.diff_texts(previous_config, latest_config,
                                        {'F': 0.1, 'ratio_mode': 'accurate', 'fast_match': True})
        # THIS IS BROKEN
        # diffs = xmldiff_main.diff_texts(previous_config, latest_config)
        # returned diffs have the following basic structure
        # InsertNode(target='/config/shared[1]', tag='log-settings', position=2)
        # InsertNode(target='/config/shared/log-settings[1]', tag='http', position=0)
        # keep a list of found xpaths
        fx = list()

        # also track nodes with updated text
        updated_text_snippets = list()

        snippets = list()
        # keep a dict of targets to xpaths
        xpaths = dict()

        # convert the config string to an xml doc
        latest_doc = ElementTree.fromstring(latest_config)

        for d in diffs:
            logger.debug(d)
            # step 1 - find all inserted nodes (future enhancement can consider other types of detected changes as well

            if 'InsertNode' in str(d):

                if d.target not in xpaths:
                    d_xpath = self.__normalize_xpath(latest_doc, d.target)
                    xpaths[d.target] = d_xpath

                else:
                    d_xpath = xpaths[d.target]

                # we have an inserted node, step2 determine if it's a top level element or a child of another element
                # xmldiff will return even inserted nodes in elements that have already been inserted
                # for purposes of building a skillet, we only need the top most unique element
                # d_target = re.sub(r'\[\d+\]$', '', d.target)
                # d_full = f'{d_target}/{d.tag}'
                # has this element been found to be a child of another element?
                found = False
                # iter all diffs again to verify if this element is actually a child element or a top level element

                for e in diffs:
                    # again only consider inserted nodes for now
                    if 'InsertNode' in str(e):

                        if e.target not in xpaths:
                            e_xpath = self.__normalize_xpath(latest_doc, e.target)
                            xpaths[e.target] = e_xpath

                        else:
                            e_xpath = xpaths[e.target]

                        e_target = re.sub(r'\[\d+\]$', '', e.target)
                        e_full = f'{e_target}/{e.tag}'
                        # begin checking for child / parent or equality relationship

                        if e.target == d.target and d.tag == e.tag and d.position == e.position:
                            # this is the same diff
                            pass

                        elif e.target == d.target and d.tag == e.tag and d.position != e.position:
                            # new tag under an existing one possible
                            pass

                        elif e.target == d.target and d.tag == e.tag:
                            # same target and same tag indicate another entry under the same top-level tag
                            # do not keep this as a top-level element
                            found = True
                            logger.debug('same target, same tag')
                            logger.debug(e)
                            logger.debug('---')
                            break

                        elif e.target != d.target and e_full in d.target:
                            # the targets are not the same and this diffs target is found to be 'in' the outer diff
                            # target, therefore this cannot be a top level element
                            found = True
                            logger.debug('e_full in d.target')
                            logger.debug(e_full)
                            logger.debug('---')
                            break

                if not found:
                    # we have not found this to be a child or peer of another element
                    # therefore this must be a top-level element, let's keep it for future work
                    logger.debug(f'Appending {d} to list of changes')
                    fx.append(d)

            elif 'UpdateTextIn' in str(d):
                snippet = dict()
                normalized_xpath = self.__normalize_xpath(latest_doc, d.node)
                xpath_parts = normalized_xpath.split('/')
                xpath = "/".join(xpath_parts[:-1])
                tag = xpath_parts[-1]
                relative_xpath = re.sub(r'^\./', '/config/', xpath)
                random_name = str(int(random.random() * 1000000))
                snippet['name'] = f'{tag}-{random_name}'
                snippet['xpath'] = relative_xpath
                snippet['element'] = f'<{tag}>{d.text}</{tag}>'
                updated_text_snippets.append(snippet)

        # we have found changes in the latest_config
        if fx:
            # now iterate only the top-level diffs (insertednodes only at this time)
            for f in fx:
                # target contains the full xpath, since we have the 'config' element already in 'latest_config'
                # we need to adjust the xpath to be relative. Also attach the 'tag' to the end of the xpath and
                # account for position if supplied
                # f_target_str = xpaths[f.target]

                f_tag = f.tag

                if hasattr(f, 'position'):
                    f_tag = f'{f.tag}[{f.position}]'

                if f.target in xpaths:
                    f_target_str = xpaths[f.target]

                else:
                    f_target_str = self.__normalize_xpath(latest_doc, f.target)
                    xpaths[f.target] = f_target_str

                f_target_str_relative = re.sub(r'^\./', '/config/', f_target_str)
                changed_short_xpath = f'{f_target_str}/{f_tag}'
                # get this element from the latest config xml document
                changed_element_dirty = latest_doc.find(changed_short_xpath)
                changed_element = self.__clean_uuid(changed_element_dirty)
                # keep a string of changes
                xml_string = ''
                # we can't just dump out the changed element because it'll contain the 'tag' as the outermost tag
                # so, find all the children of this 'tag' and append them to the xml_string

                for child_element in changed_element.findall('./'):
                    xml_string += ElementTree.tostring(child_element).decode(encoding='UTF-8')

                if xml_string == '':
                    # if changed_element.text:
                    #     xml_string = changed_element.text
                    # else:
                    # this is a text only node, we should catch this later with a updateTextIn diff
                    logger.debug('****************')
                    logger.debug(f'skipping {f_target_str_relative}/{f_tag} as this looks like a text only node')
                    logger.debug('****************')
                    continue

                snippet = dict()
                random_name = str(int(random.random() * 1000000))
                snippet['name'] = f'{f.tag}-{random_name}'
                snippet['xpath'] = f'{f_target_str_relative}/{f_tag}'
                snippet['element'] = xml_string.strip()
                snippet['from_insert'] = True
                # now print out to the end user
                snippets.append(snippet)

        text_update_snippets_to_include = list()
        if updated_text_snippets:

            for ut_snippet in updated_text_snippets:
                found = False

                for snippet in snippets:

                    if snippet['xpath'] in ut_snippet['xpath']:
                        # logger.debug('This text snippet is a child of one already included')
                        found = True
                        break

                if not found:
                    text_update_snippets_to_include.append(ut_snippet)

        snippets.extend(text_update_snippets_to_include)

        return snippets

    @staticmethod
    def __clean_uuid(changed_element: Element) -> Element:
        """
        Some rules and other elements contain the 'uuid' attribute. These should be removed before
        they can be applied to another device / fw. This function descends to all child nodes and removes the uuid
        attribute if found
        :param changed_element: Element in which to search
        :return: Element with all uuid attributes removed
        """
        if changed_element is None:
            return changed_element

        # check if 'uuid' is an attribute on this element
        if 'uuid' in changed_element.attrib:
            changed_element.attrib.pop('uuid')

        # this will find all child nodes with a 'uuid' attribute
        child_nodes = changed_element.findall('.//*[@uuid]')

        if child_nodes is None or len(child_nodes) == 0:
            return changed_element

        for child in child_nodes:
            child.attrib.pop('uuid')

        return changed_element

    @staticmethod
    def __normalize_xpath(document: Element, xpath: str) -> str:
        """
        create an xpath with all attributes included. The xpaths generated from the diffing library
        are guaranteed to be unique and valid against this configuration file, however, they are not
        guaranteed to be portable to another configuration. To avoid this, iterate over the xpath, search
        the document for each part of the xpath and grab any attributes and add them to the resulting
        xpath before returning it.
        :param document: ElementTree.Element that represents the configuration from which the diff were produced
        :param xpath: the xpath of the node in question, which may be indexed and have no attributes included
        :return: the fully normalized xpath which includes all the attributes included and the indexes removed
        """
        # Example xpath: /config/mgt-config/users/entry/phash[1]
        # the xpath will be absolute, change it here to be relative so we can search the document
        relative_xpath = re.sub('^/config/', './', xpath)
        # split the xpath into it's parts
        parts = relative_xpath.split('/')
        # xpath is now: ['.', 'mgt-config', 'users', 'entry', 'phash[1]']
        # begin constructing the new partial xpath. We will iteratively add additional parts, checking each one for
        # attributes that should be added
        path = ''

        for p in parts:

            if p == '.':
                # skip checking the root node, don't care about attributes here in the xpath
                path = p
                continue
            # add the next part to the previous, adding the '/'
            # example xpath: ./mgt-config/users/entry
            path = path + '/' + p
            logger.debug(f'Checking path: {path}')
            el = document.find(path)

            if el is None:
                # this should never happen as the xpath was found in the document we are checking
                # this would indicate a programmatic error somewhere
                raise PanoplyException('Could not normalize xpath in configuration document!')

            if el.attrib != {} and type(el.attrib) is dict:
                logger.debug('Found attributes here')
                # begin assembling the attributes into a string
                # resulting xpath will be something like entry[@name='rule1']
                attrib_str = ''

                for k, v in el.attrib.items():
                    attrib_str += f'[@{k}="{v}"]'

                if re.match(r'.*\[\d+\]$', path):
                    logger.debug('replacing indexed element with attribute named')
                    path = re.sub(r'\[\d+\]$', f'{attrib_str}', path)

                else:
                    path = path + attrib_str

                # example xpath is now: ./mgt-config/users/entry[@name="admin"]/phash[1]

            else:
                path = re.sub(r'\[\d+\]', '', path)
                # now removing the index from items that do not have attributes
                # example xpath now: ./mgt-config/users/entry[@name="admin"]/phash

        logger.debug(f'returning {path}')

        return path

    @staticmethod
    def get_ordered_xpaths() -> tuple:
        """
        Returns a list of ordered xpaths for use in ordering snippets and set commands

        Will be enhanced one day with version and model specific information if necessary

        :return: tuple of two lists, xpaths and post_xpaths
        """
        xpaths = [
            'shared/certificate',
            'shared/ssl-tls-service-profile',
            'shared/tag',
            'shared/profiles',
            'shared/reports',
            'shared/',  # catch the rest of the shared items here
            'tag/',
            'deviceconfig/system',
            'network/profiles',
            'network/interface',
            'network/virtual-wire',
            'network/vlan',
            'network/ike',
            'network/tunnel',
            'import/network/interface',  # for use with panorama plugin import (SD-WAN)
            'network/virtual-router',
            'network/profiles/zone-protection-profile',
            'routing-table/ip/static-route/entry/next-hop',  # try to keep next-hop before path-monitor for set cli
            'routing-table/ip/static-route/entry/path-monitor',  # #70
            'dynamic-ip-and-port/interface-address',  # GH #70
            'dynamic-ip-and-port/interface',  # GH #70
            'dynamic-ip-and-port/ip',  # GH #70
            'zone/',
            'profiles/custom-url-category',  # should come before profiles/url-filtering
            'address/'  # should come before rules or address-group
        ]

        post_xpaths = ['rulebase']

        return xpaths, post_xpaths


class EphemeralPanos(Panoply):
    """
    EphemeralPanos is used when a Panos instance may or may not be online and available at all times. This is useful
    when instantiating instances via a CI/CD pipeline or other automation tool.
    """
    pass


class Panos(Panoply):
    """
    Panos is used to connect to PAN-OS devices that are expected to be currently and always online! Exceptions
    will be raised in any event that we cannot connect!
    """

    def __init__(self, hostname: Optional[str], api_username: Optional[str], api_password: Optional[str],
                 api_port: Optional[int] = 443, serial_number: Optional[str] = None,
                 debug: Optional[bool] = False, api_key: Optional[str] = None):

        super().__init__(hostname, api_username, api_password, api_port, serial_number, debug, api_key)

        if self.connected:
            return

        if self.xapi and 'URLError' in self.xapi.status_detail:
            raise TargetConnectionException(self.xapi.status_detail)



import json
import logging
from typing import Tuple
from urllib.parse import quote

import urllib3
from requests import Response
from requests import Session

from .template import TemplateSnippet
from ..exceptions import SkilletLoaderException

logger = logging.getLogger(__name__)
urllib3.disable_warnings()


class RestSnippet(TemplateSnippet):
    """
    Rest Snippet
    """
    # required metadata items
    required_metadata = {'name'}

    name = ''
    path = ''

    template_metadata = {'path', 'element', 'headers'}

    output_type = 'rest'
    # optional metadata items and their default values
    optional_metadata = {
        'path': '',
        'operation': 'get',
        'payload': '',
        'headers': {},
        'content_type': '',
        'accepts_type': ''
    }

    operation = 'get'
    payload = ''
    headers = dict()
    content_type = ''
    accepts_type = ''

    def __init__(self, payload_str: str, metadata: dict, session: Session):

        # use element loaded in metadata if found, otherwise use passed in payload_str
        super().__init__(metadata.get('element', payload_str), metadata)
        # keep track of session from the parent skillet
        self.session = session

        if self.content_type != '':
            self.headers['Content-Type'] = self.content_type

        if self.accepts_type != '':
            self.headers['Accepts-Type'] = self.accepts_type

    def sanitize_metadata(self, metadata: dict) -> dict:
        """
        Clean and sanitize metadata elements in this snippet definition

        :param metadata: dict
        :return: dict
        """
        metadata = super().sanitize_metadata(metadata)

        # FIX for #59 - ensure operation is always lower cased
        if 'operation' in metadata:
            metadata['operation'] = str(metadata['operation']).lower()

        if metadata['operation'] not in ('post', 'get'):
            err = 'Supported operations are currently post and get only'
            raise SkilletLoaderException(f'Invalid metadata configuration: {err}')

        if 'path' in metadata:
            metadata['path'] = str(metadata['path']).strip().replace('\n', '')

        return metadata

    def execute(self, raw_context: dict) -> Tuple[str, str]:

        context = dict()

        if raw_context is not None:
            # always enforce quotes in the context
            for k, v in raw_context.items():
                if isinstance(v, str):
                    context[k] = quote(v)

        url = self.metadata['path']

        if self.operation == 'post':
            rendered_payload = self.metadata['element']
            if 'form' in self.headers.get('Content-Type', ''):
                payload = json.loads(rendered_payload)
            else:
                payload = rendered_payload

            response = self.session.post(url, data=payload, headers=self.metadata['headers'], verify=False)
            return self.__handle_response(response)

        # support 'noop' for #100
        elif self.operation == 'noop':
            return 'noop', 'success'

        else:
            # FIX for #59 - Ensure we pass headers to get operations properly
            response = self.session.get(url, verify=False, headers=self.metadata['headers'])
            return self.__handle_response(response)

    def __handle_response(self, response: Response) -> Tuple[str, str]:
        if response.status_code != 200:
            logger.error(f'Failed to execute REST snippet {self.name}: {response.status_code}')
            return response.text, 'failure'
        else:
            if 'json' in response.headers.get('content-type', ''):
                r = response.json()
            else:
                r = response.text

            return r, 'success'
